# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import functools
import logging
import os
import re
import subprocess
import sys
import types
import typing as t
from abc import ABC
from abc import abstractmethod

import attr
import inflection
import orjson

import bentoml
import onediffusion
from bentoml._internal.models.model import ModelSignature
from bentoml._internal.types import ModelSignatureDict

from .exceptions import ForbiddenAttributeError
from .exceptions import GpuNotAvailableError
from .exceptions import OneDiffusionException
from .utils import DEBUG
from .utils import LazyLoader
from .utils import ModelEnv
from .utils import bentoml_cattr
from .utils import first_not_none
from .utils import get_debug_mode
from .utils import is_torch_available
from .utils import non_intrusive_setattr
from .utils import pkg


if t.TYPE_CHECKING:
    import torch

    from bentoml._internal.runner.strategy import Strategy

    from .models.auto.factory import _BaseAutoSDClass

    class SDRunner(bentoml.Runner):
        __doc__: str
        __module__: str
        sd: onediffusion.SD[t.Any, t.Any]
        config: onediffusion.SDConfig
        sd_type: str
        identifying_params: dict[str, t.Any]

        def __call__(self, *args: t.Any, **attrs: t.Any) -> t.Any:
            ...

else:
    SDRunner = bentoml.Runner
    torch = LazyLoader("torch", globals(), "torch")

logger = logging.getLogger(__name__)


def import_model():
    pass


def convert_diffusion_model_name(name: str | None) -> str:
    if name is None:
        raise ValueError("'name' cannot be None")
    if os.path.exists(os.path.dirname(name)):
        name = os.path.basename(name)
        logger.debug("Given name is a path, only returning the basename %s")
        return name
    return re.sub("[^a-zA-Z0-9]+", "-", name)


_reserved_namespace = {"config_class", "model", "import_kwargs"}

class SDInterface(ABC):
    """This defines the loose contract for all onediffusion.SD implementations."""

    config_class: type[onediffusion.SDConfig]
    """The config class to use for this diffusion model. If you are creating a custom SD, you must specify this class."""

    def sd_post_init(self):
        """This function can be implemented if you need to initialized any additional variables that doesn't
        concern OneDiffusion internals.
        """
        pass

    @property
    def import_kwargs(self) -> tuple[dict[str, t.Any], dict[str, t.Any]] | None:
        """The default import kwargs to used when importing the model.
        This will be passed into 'openllm.LLM.import_model'.
        It returns two dictionaries: one for model kwargs and one for tokenizer kwargs.
        """
        return

    def import_model(
        self, model_id: str, tag: bentoml.Tag, *args: t.Any, **attrs: t.Any
    ) -> bentoml.Model:
        """This function can be implemented if default import_model doesn't satisfy your needs."""
        raise NotImplementedError


_M = t.TypeVar("_M")
_T = t.TypeVar("_T")


@attr.define(slots=True, repr=False)
class SD(SDInterface, t.Generic[_M, _T]):
    if t.TYPE_CHECKING:
        # The following will be populated by metaclass
        __sd_implementation__: t.Literal["pt"]
        __sd_model__: _M | None
        __sd_tag__: bentoml.Tag | None
        __sd_bentomodel__: bentoml.Model | None

        __sd_post_init__: t.Callable[[t.Self], None] | None
        __sd_custom_load__: t.Callable[[t.Self, t.Any, t.Any], None] | None
        __sd_init_kwargs__: property | None

        _model_args: tuple[t.Any, ...]
        _model_attrs: dict[str, t.Any]

        bettertransformer: bool

    def __init_subclass__(cls):
        cd = cls.__dict__
        prefix_class_name_config = cls.__name__
        implementation = "pt"

        cls.__sd_implementation__ = implementation
        config_class = onediffusion.AutoConfig.infer_class_from_name(prefix_class_name_config)

        if "__sdserver_internal__" in cd:
            if "config_class" not in cd:
                cls.config_class = config_class
            else:
                logger.debug(f"Using config class {cd['config_class']} for {cls.__name__}.")
        else:
            if "config_class" not in cd:
                raise RuntimeError(
                    "Missing required key 'config_class'. Make sure to define it within the SD subclass."
                )

        if cls.import_model is SDInterface.import_model:
            # using the default import model
            setattr(cls, "import_model", import_model)
        else:
            logger.debug("Custom 'import_model' will be used when loading model %s", cls.__name__)

        cls.__sd_post_init__ = None if cls.sd_post_init is SDInterface.sd_post_init else cls.sd_post_init
        cls.__sd_init_kwargs__ = None if cls.import_kwargs is SDInterface.import_kwargs else cls.import_kwargs

        for attr in {"bentomodel", "tag", "model"}:
            setattr(cls, f"__sd_{attr}__", None)

    # The following is the similar interface to HuggingFace pretrained protocol.
    @classmethod
    def from_pretrained(
        cls,
        model_id: str | None = None,
        pipeline: str | None = None,
        sd_config: onediffusion.SDConfig | None = None,
        *args: t.Any,
        **attrs: t.Any,
    ) -> SD[_M, _T]:
        return cls(
            model_id=model_id,
            pipeline=pipeline,
            sd_config=sd_config,
            *args,
            **attrs,
        )

    def __init__(
        self,
        model_id: str | None = None,
        pipeline: str | None = None,
        sd_config: onediffusion.SDConfig | None = None,
        *args: t.Any,
        **attrs: t.Any,
    ):
        """Initialize the SD with given pretrained model.

        Note:
        - *args to be passed to the model.
        - **attrs will first be parsed to the AutoConfig, then the rest will be parsed to the import_model
        - for tokenizer kwargs, it should be prefixed with _tokenizer_*

        For custom pretrained path, it is recommended to pass in 'onediffusion_model_version' alongside with the path
        to ensure that it won't be loaded multiple times.
        Internally, if a pretrained is given as a HuggingFace repository path , OneDiffusion will usethe commit_hash
        to generate the model version.

        For better consistency, we recommend users to also push the fine-tuned model to HuggingFace repository.

        If you need to overwrite the default ``import_model``, implement the following in your subclass:

        ```python
        def import_model(
            self,
            model_id: str,
            tag: bentoml.Tag,
            *args: t.Any,
            tokenizer_kwds: dict[str, t.Any],
            **attrs: t.Any,
        ):
            return bentoml.transformers.save_model(
                tag,
                model,
            )
        ```

        If your import model doesn't require customization, you can simply pass in `import_kwargs`
        at class level that will be then passed into The default `import_model` implementation.
        See ``bentoml.diffusers_simple.stable_diffusion`` for example.

        ```python
        stable_diffusion_runner = bentoml.diffusers_simple.create_runner(
            "stable_diffusion", torch_dtype=torch.bfloat16, device_map="gpu"
        )
        ```

        Note: If you implement your own `import_model`, then `import_kwargs` will be the
        default kwargs for every load.

        Note that this tag will be generated based on `self.default_id` or the given `pretrained` kwds.
        passed from the __init__ constructor.

        ``sd_post_init`` can also be implemented if you need to do any additional
        initialization after everything is setup.
        """

        sdserver_model_version = attrs.pop("sdserver_model_version", None)

        # # low_cpu_mem_usage is only available for model
        # # this is helpful on system with low memory to avoid OOM
        # low_cpu_mem_usage = attrs.pop("low_cpu_mem_usage", True)

        if sd_config is not None:
            logger.debug("Using provided SDConfig to initialize SD instead of from default: %r", sd_config)
            self.config = sd_config
        else:
            self.config = self.config_class.model_construct_env(**attrs)
            # The rests of the kwargs that is not used by the config class should be stored into __sdserver_extras__.
            attrs = self.config["extras"]

        model_kwds = {}
        if self.__sd_init_kwargs__:
            if t.TYPE_CHECKING:
                # the above meta value should determine that this SD has custom kwargs
                assert self.import_kwargs
            model_kwds = self.import_kwargs
            logger.debug(
                "'%s' default kwargs for model: '%s', tokenizer: '%s'",
                self.__class__.__name__,
                model_kwds,
            )

        if model_id is None:
            model_id = os.environ.get(self.config["env"].model_id, self.config["default_id"])

        if pipeline is None:
            pipeline = os.environ.get(self.config["env"].pipeline, self.config["default_pipeline"])

        # NOTE: This is the actual given path or pretrained weight for this SD.
        assert model_id is not None
        self._model_id = model_id
        self.pipeline = pipeline

        # parsing model kwargs
        model_kwds.update({k: v for k, v in attrs.items()})


        # NOTE: Save the args and kwargs for latter load
        self._model_args = args
        self._model_attrs = model_kwds
        self._sdserver_model_version = sdserver_model_version

        if self.__sd_post_init__:
            self.sd_post_init()


    def __setattr__(self, attr: str, value: t.Any):
        if attr in _reserved_namespace:
            raise ForbiddenAttributeError(
                f"{attr} should not be set during runtime "
                f"as these value will be reflected during runtime. "
                f"Instead, you can create a custom SD subclass {self.__class__.__name__}."
            )

        super().__setattr__(attr, value)

    def __repr__(self) -> str:
        keys = {"model_id", "runner_name", "sd_type", "config"}
        return f"{self.__class__.__name__}({', '.join(f'{k}={getattr(self, k)!r}' for k in keys)})"

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def runner_name(self) -> str:
        return f"onediffusion-{self.config['start_name']}-runner"

    @property
    def sd_type(self) -> str:
        return convert_diffusion_model_name(self._model_id)

    @property
    def identifying_params(self) -> dict[str, t.Any]:
        return {
            "configuration": self.config.model_dump_json().decode(),
            "model_ids": orjson.dumps(self.config["model_ids"]).decode(),
        }





# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some imports utils are vendorred from transformers/utils/import_utils.py for performance reasons.
"""
from __future__ import annotations

import importlib
import importlib.metadata
import importlib.util
import logging
import os
import typing as t
from abc import ABCMeta
from collections import OrderedDict

import inflection
from packaging import version

from bentoml._internal.utils import LazyLoader
from bentoml._internal.utils import pkg


if t.TYPE_CHECKING:
    BackendOrderredDict = OrderedDict[str, tuple[t.Callable[[], bool], str]]
else:
    BackendOrderredDict = OrderedDict

logger = logging.getLogger(__name__)

OPTIONAL_DEPENDENCIES = {"fine-tune", "flan-t5", "openai", "agents"}
ENV_VARS_TRUE_VALUES = {"1", "ON", "YES", "TRUE"}
ENV_VARS_TRUE_AND_AUTO_VALUES = ENV_VARS_TRUE_VALUES.union({"AUTO"})

USE_TF = os.environ.get("USE_TF", "AUTO").upper()
USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()
USE_JAX = os.environ.get("USE_FLAX", "AUTO").upper()

FORCE_TF_AVAILABLE = os.environ.get("FORCE_TF_AVAILABLE", "AUTO").upper()


def _is_package_available(package: str) -> bool:
    _package_available = importlib.util.find_spec(package) is not None
    if _package_available:
        try:
            importlib.metadata.version(package)
        except importlib.metadata.PackageNotFoundError:
            _package_available = False
    return _package_available


_torch_available = importlib.util.find_spec("torch") is not None
_tf_available = importlib.util.find_spec("tensorflow") is not None
_flax_available = importlib.util.find_spec("jax") is not None and importlib.util.find_spec("flax") is not None
_einops_available = _is_package_available("einops")
_cpm_kernel_available = _is_package_available("cpm_kernels")
_bitsandbytes_available = _is_package_available("bitsandbytes")


def is_torch_available():
    global _torch_available
    if USE_TORCH in ENV_VARS_TRUE_AND_AUTO_VALUES and USE_TF not in ENV_VARS_TRUE_VALUES:
        if _torch_available:
            try:
                importlib.metadata.version("torch")
            except importlib.metadata.PackageNotFoundError:
                _torch_available = False
    else:
        logger.info("Disabling PyTorch because USE_TF is set")
        _torch_available = False
    return _torch_available


def is_flax_available():
    global _flax_available
    if USE_JAX in ENV_VARS_TRUE_AND_AUTO_VALUES:
        if _flax_available:
            try:
                importlib.metadata.version("jax")
                importlib.metadata.version("flax")
            except importlib.metadata.PackageNotFoundError:
                _flax_available = False
    else:
        _flax_available = False
    return _flax_available


PYTORCH_IMPORT_ERROR_WITH_TF = """\
{0} requires the PyTorch library but it was not found in your environment.
However, we were able to find a TensorFlow installation. TensorFlow classes begin
with "TF", but are otherwise identically named to the PyTorch classes. This
means that the TF equivalent of the class you tried to import would be "TF{0}".
If you want to use TensorFlow, please use TF classes instead!

If you really do want to use PyTorch please go to
https://pytorch.org/get-started/locally/ and follow the instructions that
match your environment.
"""

TF_IMPORT_ERROR_WITH_PYTORCH = """\
{0} requires the TensorFlow library but it was not found in your environment.
However, we were able to find a PyTorch installation. PyTorch classes do not begin
with "TF", but are otherwise identically named to our TF classes.
If you want to use PyTorch, please use those classes instead!

If you really do want to use TensorFlow, please follow the instructions on the
installation page https://www.tensorflow.org/install that match your environment.
"""

TENSORFLOW_IMPORT_ERROR = """{0} requires the TensorFlow library but it was not found in your environment.
Checkout the instructions on the installation page: https://www.tensorflow.org/install and follow the
ones that match your environment. Please note that you may need to restart your runtime after installation.
"""


FLAX_IMPORT_ERROR = """{0} requires the FLAX library but it was not found in your environment.
Checkout the instructions on the installation page: https://github.com/google/flax and follow the
ones that match your environment. Please note that you may need to restart your runtime after installation.
"""

PYTORCH_IMPORT_ERROR = """{0} requires the PyTorch library but it was not found in your environment.
Checkout the instructions on the installation page: https://pytorch.org/get-started/locally/ and follow the
ones that match your environment. Please note that you may need to restart your runtime after installation.
"""

CPM_KERNELS_IMPORT_ERROR = """{0} requires the cpm_kernels library but it was not found in your environment.
You can install it with pip: `pip install cpm_kernels`. Please note that you may need to restart your
runtime after installation.
"""

EINOPS_IMPORT_ERROR = """{0} requires the einops library but it was not found in your environment.
You can install it with pip: `pip install einops`. Please note that you may need to restart
your runtime after installation.
"""

BACKENDS_MAPPING = BackendOrderredDict(
    [
        ("torch", (is_torch_available, PYTORCH_IMPORT_ERROR)),
    ]
)


class DummyMetaclass(ABCMeta):
    """Metaclass for dummy object. It will raises ImportError
    generated by ``require_backends`` if users try to access attributes from given class
    """

    _backends: t.List[str]

    def __getattribute__(cls, key: str) -> t.Any:
        if key.startswith("_"):
            return super().__getattribute__(key)
        require_backends(cls, cls._backends)


def require_backends(o: t.Any, backends: t.MutableSequence[str]):
    if not isinstance(backends, (list, tuple)):
        backends = list(backends)

    name = o.__name__ if hasattr(o, "__name__") else o.__class__.__name__

    # Raise an error for users who might not realize that classes without "TF" are torch-only
    if "torch" in backends and not is_torch_available():
        raise ImportError(PYTORCH_IMPORT_ERROR_WITH_TF.format(name))

    checks = (BACKENDS_MAPPING[backend] for backend in backends)
    failed = [msg.format(name) for available, msg in checks if not available()]
    if failed:
        raise ImportError("".join(failed))


class ModelEnv:
    model_name: str

    if t.TYPE_CHECKING:
        config: property
        model_id: property
        pipeline: property
        lora_weights: property
        lora_dir: property
        framework: property

        framework_value: property

    def __getitem__(self, item: str | t.Any) -> t.Any:
        if hasattr(self, item):
            return getattr(self, item)
        raise KeyError(f"Key {item} not found in {self}")

    def __new__(cls, model_name: str):
        from .._configuration import _field_env_key
        from . import codegen

        model_name = inflection.underscore(model_name)

        res = super().__new__(cls)
        res.model_name = model_name

        # gen properties env key
        attributes = {"config", "model_id", "pipeline", "lora_weights", "lora_dir", "framework"}
        for att in attributes:
            setattr(res, att, _field_env_key(model_name, att.upper()))

        # gen properties env value
        attributes_with_values = {
            "framework": (str, "pt"),
        }
        globs: dict[str, t.Any] = {
            "__bool_vars_value": ENV_VARS_TRUE_VALUES,
            "__env_get": os.environ.get,
            "self": res,
        }

        for attribute, (default_type, default_value) in attributes_with_values.items():
            lines: list[str] = []
            if default_type is bool:
                lines.append(
                    f"return str(__env_get(self['{attribute}'], str(__env_default)).upper() in __bool_vars_value)"
                )
            else:
                lines.append(f"return __env_get(self['{attribute}'], __env_default)")

            setattr(
                res,
                f"{attribute}_value",
                codegen.generate_function(
                    cls,
                    "_env_get_" + attribute,
                    lines,
                    ("__env_default",),
                    globs,
                )(default_value),
            )

        return res

    @property
    def start_docstring(self) -> str:
        return getattr(self.module, f"START_{self.model_name.upper()}_COMMAND_DOCSTRING")

    @property
    def module(self) -> LazyLoader:
        return LazyLoader(self.model_name, globals(), f"onediffusion.models.{self.model_name}")



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import typing as t
from collections import OrderedDict

from .configuration_auto import CONFIG_MAPPING_NAMES
from .factory import _BaseAutoSDClass
from .factory import _LazyAutoMapping


if t.TYPE_CHECKING:
    import onediffusion

MODEL_MAPPING_NAMES = OrderedDict(
    [
        ("stable_diffusion", "StableDiffusion"),
        ("stable_diffusion_xl", "StableDiffusionXL"),
    ]
)

MODEL_MAPPING: dict[
    type[onediffusion.SDConfig], type[onediffusion.SD[t.Any, t.Any]]
] = _LazyAutoMapping(CONFIG_MAPPING_NAMES, MODEL_MAPPING_NAMES)


class AutoSD(_BaseAutoSDClass):
    _model_mapping = MODEL_MAPPING



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
OneDiffusion
============

An open platform for operating diffusion models in production. Fine-tune, serve,
deploy, and monitor any diffusion models with ease.

* Built-in support for Stable Diffusion, ControlNet
* Option to bring your own fine-tuned diffusion models
* Online Serving with HTTP
* Native integration with BentoML
"""
from __future__ import annotations

import logging
import typing as t

from . import utils as utils
from .__about__ import __version__ as __version__
from .exceptions import MissingDependencyError


if utils.DEBUG:
    utils.set_debug_mode(True)
    utils.set_quiet_mode(False)

    utils.configure_logging()
    logging.basicConfig(level=logging.NOTSET)


_import_structure = {
    "_sd": ["SD"],
    "_configuration": ["SDConfig"],
    "_package": ["build"],
    "exceptions": [],
    "utils": [],
    "models": [],
    "cli": ["start"],
    # NOTE: models
    "models.auto": [
        "AutoConfig",
        "CONFIG_MAPPING",
        "MODEL_MAPPING_NAMES",
        # for flax support later
        # "MODEL_FLAX_MAPPING_NAMES",
    ],
    "models.stable_diffusion": ["StableDiffusionConfig"],
}

try:
    if not utils.is_torch_available():
        raise MissingDependencyError
except MissingDependencyError:
    from .utils import dummy_pt_objects

    _import_structure["utils.dummy_pt_objects"] = [name for name in dir(dummy_pt_objects) if not name.startswith("_")]
else:
    # _import_structure["models.stable_diffusion"].extend(["StableDiffusion"])
    _import_structure["models.auto"].extend(["AutoSD", "MODEL_MAPPING"])


# declaration for OneDiffusion-related modules
if t.TYPE_CHECKING:
    from . import cli as cli
    # from . import client as client
    from . import exceptions as exceptions
    from . import models as models

    # Specific types import
    from ._configuration import SDConfig as SDConfig
    from .models.auto import CONFIG_MAPPING as CONFIG_MAPPING
    from .models.auto import AutoConfig as AutoConfig

    try:
        if not utils.is_torch_available():
            raise MissingDependencyError
    except MissingDependencyError:
        from .utils.dummy_pt_objects import *
    else:
        from .models.auto import MODEL_MAPPING as MODEL_MAPPING
        from .models.auto import AutoSD as AutoSD

else:
    import sys

    sys.modules[__name__] = utils.LazyModule(
        __name__,
        globals()["__file__"],
        _import_structure,
        module_spec=__spec__,
        extra_objects={
            "__version__": __version__,
            # The below is a special mapping that allows OneDIffusion to
            # be used as a dictionary.  This is purely for convenience
            # sake, and should not be used in performance critcal
            # code. This is also not considered as a public API.

            "__sdserver_special__": {"pt": "AutoSD"},
        },
    )



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import string
import typing as t
from pathlib import Path


if t.TYPE_CHECKING:
    from fs.base import FS

    DictStrAny = dict[str, t.Any]

    from attr import _make_method
else:
    # NOTE: Using internal API from attr here, since we are actually
    # allowing subclass of openllm.LLMConfig to become 'attrs'-ish
    from attr._make import _make_method

    DictStrAny = dict

_T = t.TypeVar("_T")

logger = logging.getLogger(__name__)

OPENLLM_MODEL_NAME = "# openllm: model name"
OPENLLM_MODEL_ID = "# openllm: model id"


class ModelNameFormatter(string.Formatter):
    model_keyword: t.LiteralString = "__model_name__"

    def __init__(self, model_name: str):
        super().__init__()
        self.model_name = model_name

    def vformat(self, format_string: str) -> str:
        return super().vformat(format_string, (), {self.model_keyword: self.model_name})

    def can_format(self, value: str) -> bool:
        try:
            self.parse(value)
            return True
        except ValueError:
            return False


class ModelIdFormatter(ModelNameFormatter):
    model_keyword: t.LiteralString = "__model_id__"


_service_file = Path(__file__).parent.parent / "_service.py"


def write_service(model_name: str, model_id: str, target_path: str, llm_fs: FS):
    from . import DEBUG

    logger.debug("Generating service for %s to %s", model_name, target_path)
    with open(_service_file.__fspath__(), "r") as f:
        src_contents = f.readlines()

    # modify with model name
    for it in src_contents:
        if OPENLLM_MODEL_NAME in it:
            src_contents[src_contents.index(it)] = ModelNameFormatter(model_name).vformat(it)
        elif OPENLLM_MODEL_ID in it:
            src_contents[src_contents.index(it)] = ModelIdFormatter(model_id).vformat(it)

    script = f"# GENERATED BY 'openllm build {model_name}'. DO NOT EDIT\n" + "".join(src_contents)

    if DEBUG:
        logger.info("Generated script:\n%s", script)

    llm_fs.writetext(target_path, script)


# NOTE: The following ins extracted from attrs internal APIs

# sentinel object for unequivocal object() getattr
_sentinel = object()


def has_own_attribute(cls: type[t.Any], attrib_name: t.Any):
    """
    Check whether *cls* defines *attrib_name* (and doesn't just inherit it).
    """
    attr = getattr(cls, attrib_name, _sentinel)
    if attr is _sentinel:
        return False

    for base_cls in cls.__mro__[1:]:
        a = getattr(base_cls, attrib_name, None)
        if attr is a:
            return False

    return True


def get_annotations(cls: type[t.Any]) -> DictStrAny:
    """
    Get annotations for *cls*.
    """
    if has_own_attribute(cls, "__annotations__"):
        return cls.__annotations__

    return DictStrAny()


_classvar_prefixes = (
    "typing.ClassVar",
    "t.ClassVar",
    "ClassVar",
    "typing_extensions.ClassVar",
)


def is_class_var(annot: str | t.Any) -> bool:
    """
    Check whether *annot* is a typing.ClassVar.

    The string comparison hack is used to avoid evaluating all string
    annotations which would put attrs-based classes at a performance
    disadvantage compared to plain old classes.
    """
    annot = str(annot)

    # Annotation can be quoted.
    if annot.startswith(("'", '"')) and annot.endswith(("'", '"')):
        annot = annot[1:-1]

    return annot.startswith(_classvar_prefixes)


def add_method_dunders(cls: type[t.Any], method_or_cls: _T, _overwrite_doc: str | None = None) -> _T:
    """
    Add __module__ and __qualname__ to a *method* if possible.
    """
    try:
        method_or_cls.__module__ = cls.__module__
    except AttributeError:
        pass

    try:
        method_or_cls.__qualname__ = ".".join((cls.__qualname__, method_or_cls.__name__))
    except AttributeError:
        pass

    try:
        method_or_cls.__doc__ = (
            _overwrite_doc or "Method or class generated by LLMConfig for class " f"{cls.__qualname__}."
        )
    except AttributeError:
        pass

    return method_or_cls


def generate_unique_filename(cls: type[t.Any], func_name: str):
    return f"<{cls.__name__} generated {func_name} {cls.__module__}." f"{getattr(cls, '__qualname__', cls.__name__)}>"


def generate_function(
    typ: type[t.Any],
    func_name: str,
    lines: list[str] | None,
    args: tuple[str, ...] | None,
    globs: dict[str, t.Any],
    annotations: dict[str, t.Any] | None = None,
):
    from . import DEBUG

    script = "def %s(%s):\n    %s\n" % (
        func_name,
        ", ".join(args) if args is not None else "",
        "\n    ".join(lines) if lines else "pass",
    )
    meth = _make_method(func_name, script, generate_unique_filename(typ, func_name), globs)
    if annotations:
        meth.__annotations__ = annotations

    if DEBUG:
        logger.info("Generated script for %s:\n\n%s", typ, script)

    return meth



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import typing as t

import onediffusion


_import_structure = {
    "configuration_stable_diffusion": ["StableDiffusionConfig", "START_STABLE_DIFFUSION_COMMAND_DOCSTRING",],
}

try:
    if not onediffusion.utils.is_torch_available():
        raise onediffusion.exceptions.MissingDependencyError
except onediffusion.exceptions.MissingDependencyError:
    pass
else:
    _import_structure["modeling_stable_diffusion"] = ["StableDiffusion"]
    pass

if t.TYPE_CHECKING:
    from .configuration_stable_diffusion import START_STABLE_DIFFUSION_COMMAND_DOCSTRING as START_STABLE_DIFFUSION_COMMAND_DOCSTRING
    from .configuration_stable_diffusion import StableDiffusionConfig as StableDiffusionConfig

    try:
        if not onediffusion.utils.is_torch_available():
            raise onediffusion.exceptions.MissingDependencyError
    except onediffusion.exceptions.MissingDependencyError:
        pass
    else:
        #from .modeling_dolly_v2 import DollyV2 as DollyV2
        pass
else:
    import sys

    sys.modules[__name__] = onediffusion.utils.LazyModule(
        __name__, globals()["__file__"], _import_structure, module_spec=__spec__
    )



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities function for OneDiffusion. User can import these function for convenience, but
we won't ensure backward compatibility for these functions. So use with caution.
"""
from __future__ import annotations as _annotations

import functools
import logging
import logging.config
import os
import sys
import types
import typing as t

from bentoml._internal.configuration import get_debug_mode
from bentoml._internal.configuration import get_quiet_mode
from bentoml._internal.configuration import set_debug_mode
from bentoml._internal.configuration import set_quiet_mode
from bentoml._internal.log import CLI_LOGGING_CONFIG as _CLI_LOGGING_CONFIG
from bentoml._internal.types import LazyType
from bentoml._internal.utils import LazyLoader
from bentoml._internal.utils import bentoml_cattr
from bentoml._internal.utils import copy_file_to_fs_folder
from bentoml._internal.utils import first_not_none
from bentoml._internal.utils import pkg
from bentoml._internal.utils import reserve_free_port
from bentoml._internal.utils import resolve_user_filepath

from .lazy import LazyModule


logger = logging.getLogger(__name__)

try:
    from typing import GenericAlias as _TypingGenericAlias  # type: ignore
except ImportError:
    # python < 3.9 does not have GenericAlias (list[int], tuple[str, ...] and so on)
    _TypingGenericAlias = ()

if sys.version_info < (3, 10):
    _WithArgsTypes = (_TypingGenericAlias,)
else:
    _WithArgsTypes: t.Any = (
        t._GenericAlias,  # type: ignore (_GenericAlias is the actual GenericAlias implementation)
        types.GenericAlias,
        types.UnionType,
    )


def lenient_issubclass(cls: t.Any, class_or_tuple: type[t.Any] | tuple[type[t.Any], ...] | None) -> bool:
    try:
        return isinstance(cls, type) and issubclass(cls, class_or_tuple)  # type: ignore[arg-type]
    except TypeError:
        if isinstance(cls, _WithArgsTypes):
            return False
        raise


def gpu_count() -> tuple[int, ...]:
    from bentoml._internal.resource import NvidiaGpuResource

    return tuple(NvidiaGpuResource.from_system())


# equivocal setattr to save one lookup per assignment
_object_setattr = object.__setattr__


def non_intrusive_setattr(obj: t.Any, name: str, value: t.Any) -> None:
    """This makes sure that we don't overwrite any existing attributes on the object"""
    _setattr = functools.partial(setattr, obj) if isinstance(obj, type) else _object_setattr.__get__(obj)

    if not hasattr(obj, name):
        _setattr(name, value)


DEBUG = sys.flags.dev_mode or (not sys.flags.ignore_environment and bool(os.environ.get("ONEDIFFUSIONDEVDEBUG")))


_LOGGING_CONFIG = _CLI_LOGGING_CONFIG.copy()
_LOGGING_CONFIG["loggers"].update(
    {
        "onediffusion": {
            "level": logging.INFO,
            "handlers": ["bentomlhandler", "defaulthandler"],
            "propagate": False,
        }
    }
)


def configure_logging() -> None:
    if get_quiet_mode():
        _LOGGING_CONFIG["loggers"]["onediffusion"]["level"] = logging.ERROR
        _LOGGING_CONFIG["loggers"]["bentoml"]["level"] = logging.ERROR
        _LOGGING_CONFIG["root"]["level"] = logging.ERROR
    elif get_debug_mode() or DEBUG:
        _LOGGING_CONFIG["loggers"]["onediffusion"]["level"] = logging.DEBUG
        _LOGGING_CONFIG["loggers"]["bentoml"]["level"] = logging.DEBUG
        _LOGGING_CONFIG["root"]["level"] = logging.DEBUG
    else:
        _LOGGING_CONFIG["loggers"]["onediffusion"]["level"] = logging.INFO
        _LOGGING_CONFIG["loggers"]["bentoml"]["level"] = logging.INFO
        _LOGGING_CONFIG["root"]["level"] = logging.INFO

    logging.config.dictConfig(_LOGGING_CONFIG)


# NOTE: The set marks contains a set of modules name
# that are available above and are whitelisted
# to be included in the extra_objects map.
_whitelist_modules = {"pkg"}

# XXX: define all classes, functions import above this line
# since _extras will be the locals() import from this file.
_extras: dict[str, t.Any] = {
    k: v
    for k, v in locals().items()
    if k in _whitelist_modules or (not isinstance(v, types.ModuleType) and not k.startswith("_"))
}

_import_structure = {
    "analytics": [],
    "codegen": [],
    "dantic": [],
    "import_utils": [
        "OPTIONAL_DEPENDENCIES",
        "ENV_VARS_TRUE_VALUES",
        "DummyMetaclass",
        "ModelEnv",
        "is_flax_available",
        "is_torch_available",
        "require_backends",
    ],
}

if t.TYPE_CHECKING:
    # NOTE: The following exports useful utils from bentoml
    from . import LazyLoader as LazyLoader
    from . import LazyType as LazyType
    from . import analytics as analytics
    from . import bentoml_cattr as bentoml_cattr
    from . import codegen as codegen
    from . import configure_logging as configure_logging
    from . import copy_file_to_fs_folder as copy_file_to_fs_folder
    from . import dantic as dantic
    from . import first_not_none as first_not_none
    from . import get_debug_mode as get_debug_mode
    from . import get_quiet_mode as get_quiet_mode
    from . import gpu_count as gpu_count
    from . import lenient_issubclass as lenient_issubclass
    from . import non_intrusive_setattr as non_intrusive_setattr
    from . import pkg as pkg
    from . import reserve_free_port as reserve_free_port
    from . import resolve_user_filepath as resolve_user_filepath
    from . import set_debug_mode as set_debug_mode
    from . import set_quiet_mode as set_quiet_mode
    from .import_utils import ENV_VARS_TRUE_VALUES as ENV_VARS_TRUE_VALUES
    from .import_utils import OPTIONAL_DEPENDENCIES as OPTIONAL_DEPENDENCIES
    from .import_utils import DummyMetaclass as DummyMetaclass
    from .import_utils import ModelEnv as ModelEnv
    from .import_utils import is_torch_available as is_torch_available
    from .import_utils import require_backends as require_backends
    from .lazy import LazyModule as LazyModule
else:
    import sys

    sys.modules[__name__] = LazyModule(
        __name__,
        globals()["__file__"],
        _import_structure,
        module_spec=__spec__,
        extra_objects=_extras,
    )



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Types definition for OneDiffusion.

Note that this module SHOULD NOT BE IMPORTED DURING RUNTIME, as this serve only for typing purposes.
It will raises a RuntimeError if this is imported eagerly.
"""
from __future__ import annotations

import typing as t

if not t.TYPE_CHECKING:
    raise RuntimeError(f"{__name__} should not be imported during runtime")

import click

P = t.ParamSpec("P")
O_co = t.TypeVar("O_co", covariant=True)

from bentoml._internal.models.model import ModelSignaturesType as ModelSignaturesType


class ClickFunctionWrapper(t.Protocol[P, O_co]):
    __name__: str
    __click_params__: list[click.Option]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> O_co:
        ...


# F is a t.Callable[P, O_co] with compatible to ClickFunctionWrapper
class F(t.Generic[P, O_co]):
    __name__: str
    __click_params__: list[click.Option]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> O_co:
        ...



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A shim provides usable transition from pydantic to attrs."""

from __future__ import annotations

import functools
import importlib
import os
import typing as t
from enum import Enum

import attr
import click
import click_option_group as cog
import inflection
import orjson
from click import ParamType

import onediffusion


if t.TYPE_CHECKING:
    from attr import _ValidatorType

    from .._types import ClickFunctionWrapper
    from .._types import F
    from .._types import O_co
    from .._types import P

_T = t.TypeVar("_T")


@t.overload
def attrs_to_options(
    name: str,
    field: attr.Attribute[t.Any],
    model_name: str,
    typ: type[t.Any] | None = None,
    suffix_generation: bool = False,
) -> F[..., F[..., onediffusion.SDConfig]]:
    ...


@t.overload
def attrs_to_options(  # type: ignore (overlapping overload)
    name: str,
    field: attr.Attribute[O_co],
    model_name: str,
    typ: type[t.Any] | None = None,
    suffix_generation: bool = False,
) -> F[..., F[P, O_co]]:
    ...


def attrs_to_options(
    name: str,
    field: attr.Attribute[t.Any],
    model_name: str,
    typ: type[t.Any] | None = None,
    suffix_generation: bool = False,
) -> t.Callable[..., ClickFunctionWrapper[..., t.Any]]:
    # TODO: support parsing nested attrs class and Union
    envvar = field.metadata["env"]
    dasherized = inflection.dasherize(name)
    underscored = inflection.underscore(name)

    if typ in (None, attr.NOTHING):
        typ = field.type

    full_option_name = f"--{dasherized}"
    if field.type is bool:
        full_option_name += f"/--no-{dasherized}"
    if suffix_generation:
        identifier = f"{model_name}_generation_{underscored}"
    else:
        identifier = f"{model_name}_{underscored}"

    return cog.optgroup.option(
        identifier,
        full_option_name,
        type=parse_type(typ),
        required=field.default is attr.NOTHING,
        default=field.default if field.default not in (attr.NOTHING, None) else None,
        show_default=True,
        multiple=allows_multiple(typ),
        help=field.metadata.get("description", "(No description provided)"),
        show_envvar=True,
        envvar=envvar,
    )


def env_converter(value: t.Any, env: str | None = None) -> t.Any:
    if env is not None:
        value = os.environ.get(env, value)
    if value is not None and isinstance(value, str):
        try:
            return orjson.loads(value.lower())
        except orjson.JSONDecodeError as err:
            raise RuntimeError(f"Failed to parse '{value}' from '{env}': {err}")
    return value


def Field(
    default: t.Any = None,
    *,
    ge: int | float | None = None,
    le: int | float | None = None,
    validator: _ValidatorType[_T] | None = None,
    description: str | None = None,
    env: str | None = None,
    auto_default: bool = False,
    use_default_converter: bool = True,
    **attrs: t.Any,
):
    """A decorator that extends attr.field with additional arguments, which provides the same
    interface as pydantic's Field.

    By default, if both validator and ge are provided, then then ge will be
    piped into first, then all of the other validator will be run afterwards.

    Args:
        ge: Greater than or equal to. Defaults to None.
        description: the documentation for the field. Defaults to None.
        env: the environment variable to read from. Defaults to None.
        auto_default: a bool indicating whether to use the default value as the environment.
            Defaults to False. If set to True, the behaviour of this Field will also depends
            on kw_only. If kw_only=True, the this field will become 'Required' and the default
            value is omitted. If kw_only=False, then the default value will be used as before.
        use_default_converter: a bool indicating whether to use the default converter. Defaults
            to True. If set to False, then the default converter will not be used. The default
            converter converts a given value from the environment variable for this given Field.
        **kwargs: The rest of the arguments are passed to attr.field
    """
    metadata = attrs.pop("metadata", {})
    if description is None:
        description = "(No description provided)"
    metadata["description"] = description
    if env is not None:
        metadata["env"] = env
    piped: list[_ValidatorType[t.Any]] = []

    converter = attrs.pop("converter", None)
    if use_default_converter:
        converter = functools.partial(env_converter, env=env)

    if ge is not None:
        piped.append(attr.validators.ge(ge))
    if le is not None:
        piped.append(attr.validators.le(le))
    if validator is not None:
        piped.append(validator)

    if len(piped) == 0:
        _validator = None
    elif len(piped) == 1:
        _validator = piped[0]
    else:
        _validator = attr.validators.and_(*piped)

    factory = attrs.pop("factory", None)
    if factory is not None and default is not None:
        raise RuntimeError("'factory' and 'default' are mutually exclusive.")
    # NOTE: the behaviour of this is we will respect factory over the default
    if factory is not None:
        attrs["factory"] = factory
    else:
        attrs["default"] = default

    kw_only = attrs.pop("kw_only", False)
    if auto_default and kw_only:
        attrs.pop("default")

    return attr.field(metadata=metadata, validator=_validator, converter=converter, **attrs)


def parse_type(field_type: t.Any) -> ParamType | tuple[ParamType]:
    """Transforms the pydantic field's type into a click-compatible type.

    Args:
        field_type: pydantic field type

    Returns:
        ParamType: click type equivalent
    """
    from . import lenient_issubclass

    assert t.get_origin(field_type) is not t.Union, "Unions are not supported"
    # enumeration strings or other Enum derivatives
    if lenient_issubclass(field_type, Enum):
        return EnumChoice(enum=field_type, case_sensitive=True)
    # literals are enum-like with way less functionality
    if is_literal(field_type):
        return LiteralChoice(enum=field_type, case_sensitive=True)
    # modules, classes, functions
    if is_typing(field_type):
        return ModuleType()
    # entire dictionaries:
    # using a Dict, convert in advance
    if is_mapping(field_type):
        return JsonType()
    # list, List[p], Tuple[p], Set[p] and so on
    if is_container(field_type):
        return parse_container_args(field_type)
    # bytes are not natively supported by click
    if lenient_issubclass(field_type, bytes):
        return BytesType()
    # return the current type: it should be a primitive
    return field_type


def is_typing(field_type: type) -> bool:
    """Checks whether the current type is a module-like type.

    Args:
        field_type: pydantic field type

    Returns:
        bool: true if the type is itself a type
    """
    raw = t.get_origin(field_type)
    if raw is None:
        return False
    if raw is type or raw is t.Type:
        return True
    return False


def is_literal(field_type: type) -> bool:
    """Checks whether the given field type is a Literal type or not.
    Literals are weird: isinstance and subclass do not work, so you compare
    the origin with the Literal declaration itself.

    Args:
        field_type: current pydantic type

    Returns:
        bool: true if Literal type, false otherwise
    """
    origin = t.get_origin(field_type)
    return origin is not None and origin is t.Literal


class ModuleType(ParamType):
    name = "module"

    def _import_object(self, value: str) -> t.Any:
        module_name, class_name = value.rsplit(".", maxsplit=1)
        assert all(s.isidentifier() for s in module_name.split(".")), f"'{value}' is not a valid module name"
        assert class_name.isidentifier(), f"Variable '{class_name}' is not a valid identifier"

        module = importlib.import_module(module_name)
        if class_name:
            try:
                return getattr(module, class_name)
            except AttributeError:
                raise ImportError(f"Module '{module_name}' does not define a '{class_name}' variable.")
        return None

    def convert(self, value: str, param: click.Parameter | None, ctx: click.Context | None) -> t.Any:
        try:
            if isinstance(value, str):
                return self._import_object(value)
            return value
        except Exception as exc:
            self.fail(f"'{value}' is not a valid object ({type(exc)}: {str(exc)})", param, ctx)


class EnumChoice(click.Choice):
    name = "enum"

    def __init__(self, enum: Enum, case_sensitive: bool = False):
        self.mapping = enum
        self.internal_type = enum
        super().__init__([e.name for e in self.mapping], case_sensitive)

    def convert(self, value: t.Any, param: click.Parameter | None, ctx: click.Context | None) -> Enum:
        if isinstance(value, self.internal_type):
            return value
        result = super().convert(value, param, ctx)
        if isinstance(result, str):
            result = self.mapping[result]
        return result


class LiteralChoice(EnumChoice):
    name = "literal"

    def __init__(self, enum: t.LiteralString, case_sensitive: bool = False):
        # expect every literal value to belong to the same primitive type
        values = list(enum.__args__)
        item_type = type(values[0])
        assert all(isinstance(v, item_type) for v in values), f"Field {enum} contains items of different types"
        self.internal_type = item_type
        self.mapping = {str(v): v for v in values}
        super(EnumChoice, self).__init__(list(self.mapping.keys()), case_sensitive)


def allows_multiple(field_type: t.Any) -> bool:
    """Checks whether the current type allows for multiple arguments to be provided as input or not.
    For containers, it exploits click's support for lists and such to use the same option multiple times
    to create a complex object: `python run.py --subsets train --subsets test`
    # becomes `subsets: ["train", "test"]`.
    Args:
        field_type (type): pydantic type

    Returns:
        bool: true if it's a composite field (lists, containers and so on), false otherwise
    """
    # Early out for mappings, since it's better to deal with them using strings.
    if is_mapping(field_type):
        return False
    # Activate multiple option for (simple) container types
    if is_container(field_type):
        args = parse_container_args(field_type)
        # A non-composite type has a single argument, such as 'List[int]'
        # A composite type has a tuple of arguments, like 'Tuple[str, int, int]'.
        # For the moment, only non-composite types are allowed.
        return not isinstance(args, tuple)
    return False


def is_mapping(field_type: type) -> bool:
    """Checks whether this field represents a dictionary or JSON object.

    Args:
        field_type (type): pydantic type

    Returns:
        bool: true when the field is a dict-like object, false otherwise.
    """
    # Early out for standard containers.
    if onediffusion.utils.lenient_issubclass(field_type, t.Mapping):
        return True
    # for everything else or when the typing is more complex, check its origin
    origin = t.get_origin(field_type)
    if origin is None:
        return False
    return onediffusion.utils.lenient_issubclass(origin, t.Mapping)


def is_container(field_type: type) -> bool:
    """Checks whether the current type is a container type ('contains' other types), like
    lists and tuples.

    Args:
        field_type: pydantic field type

    Returns:
        bool: true if a container, false otherwise
    """
    # do not consider strings or byte arrays as containers
    if field_type in (str, bytes):
        return False
    # Early out for standard containers: list, tuple, range
    if onediffusion.utils.lenient_issubclass(field_type, t.Container):
        return True
    origin = t.get_origin(field_type)
    # Early out for non-typing objects
    if origin is None:
        return False
    return onediffusion.utils.lenient_issubclass(origin, t.Container)


def parse_container_args(field_type: type[t.Any]) -> ParamType | tuple[ParamType]:
    """Parses the arguments inside a container type (lists, tuples and so on).

    Args:
        field_type: pydantic field type

    Returns:
        ParamType | tuple[ParamType]: single click-compatible type or a tuple
    """
    assert is_container(field_type), "Field type is not a container"
    args = t.get_args(field_type)
    # Early out for untyped containers: standard lists, tuples, List[Any]
    # Use strings when the type is unknown, avoid click's type guessing
    if len(args) == 0:
        return str
    # Early out for homogenous containers: Tuple[int], List[str]
    if len(args) == 1:
        return parse_single_arg(args[0])
    # Early out for homogenous tuples of indefinite length: Tuple[int, ...]
    if len(args) == 2 and args[1] is Ellipsis:
        return parse_single_arg(args[0])
    # Then deal with fixed-length containers: Tuple[str, int, int]
    return tuple(parse_single_arg(arg) for arg in args)


def parse_single_arg(arg: type) -> ParamType:
    """Returns the click-compatible type for container origin types.
    In this case, returns string when it's not inferrable, a JSON for mappings
    and the original type itself in every other case (ints, floats and so on).
    Bytes is a special case, not natively handled by click.

    Args:
        arg (type): single argument

    Returns:
        ParamType: click-compatible type
    """
    # When we don't know the type, we choose 'str'
    if arg is t.Any:
        return str
    # For containers and nested models, we use JSON
    if is_container(arg):
        return JsonType()
    if onediffusion.utils.lenient_issubclass(arg, bytes):
        return BytesType()
    return arg


class BytesType(ParamType):
    name = "bytes"

    def convert(self, value: t.Any, param: click.Parameter | None, ctx: click.Context | None) -> t.Any:
        if isinstance(value, bytes):
            return value
        try:
            return str.encode(value)
        except Exception as exc:
            self.fail(f"'{value}' is not a valid string ({str(exc)})", param, ctx)


class JsonType(ParamType):
    name = "json"

    def __init__(self, should_load: bool = True) -> None:
        super().__init__()
        self.should_load = should_load

    def convert(self, value: t.Any, param: click.Parameter | None, ctx: click.Context | None) -> t.Any:
        if onediffusion.utils.LazyType[t.Mapping[str, str]](t.Mapping).isinstance(value) or not self.should_load:
            return value
        try:
            return orjson.loads(value)
        except orjson.JSONDecodeError as exc:
            self.fail(f"'{value}' is not a valid JSON string ({str(exc)})", param, ctx)



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
import importlib.machinery
import itertools
import os
import types
import typing as t

from ..exceptions import ForbiddenAttributeError
from ..exceptions import OneDiffusionException


class UsageNotAllowedError(OneDiffusionException):
    """Raised when LazyModule.__getitem__ is forbidden."""


class MissingAttributesError(OneDiffusionException):
    """Raised when given keys is not available in LazyModule special mapping."""


_reserved_namespace = {"__sdserver_special__"}


class LazyModule(types.ModuleType):
    """
    Module class that surfaces all objects but only performs associated imports when the objects are requested.
    This is a direct port from transformers.utils.import_utils._LazyModule for
    backwards compatibility with transformers < 4.18

    This is an extension a more powerful LazyLoader.
    """

    # Very heavily inspired by optuna.integration._IntegrationModule
    # https://github.com/optuna/optuna/blob/master/optuna/integration/__init__.py
    def __init__(
        self,
        name: str,
        module_file: str,
        import_structure: dict[str, list[str]],
        module_spec: importlib.machinery.ModuleSpec | None = None,
        extra_objects: dict[str, t.Any] | None = None,
    ):
        super().__init__(name)
        self._modules = set(import_structure.keys())
        self._class_to_module: dict[str, str] = {}
        _extra_objects = {} if extra_objects is None else extra_objects
        for key, values in import_structure.items():
            for value in values:
                self._class_to_module[value] = key
        # Needed for autocompletion in an IDE
        self.__all__ = list(import_structure.keys()) + list(itertools.chain(*import_structure.values()))
        self.__file__ = module_file
        self.__spec__ = module_spec
        self.__path__ = [os.path.dirname(module_file)]
        self._objects = _extra_objects
        self._name = name
        self._import_structure = import_structure

    # Needed for autocompletion in an IDE
    def __dir__(self):
        result = t.cast("list[str]", super().__dir__())
        # The elements of self.__all__ that are submodules may or
        # may not be in the dir already, depending on whether
        # they have been accessed or not. So we only add the
        # elements of self.__all__ that are not already in the dir.
        for attribute in self.__all__:
            if attribute not in result:
                result.append(attribute)
        return result

    def __getitem__(self, key: str) -> t.Any:
        # currently, this is reserved to only internal uses and users shouldn't use this.
        if self._objects.get("__sdserver_special__") is None:
            raise UsageNotAllowedError(f"'{self._name}' is not allowed to be used as a dict.")
        _special_mapping = self._objects.get("__sdserver_special__", {})
        try:
            if key in _special_mapping:
                return getattr(self, _special_mapping.__getitem__(key))
            raise MissingAttributesError(f"Requested '{key}' is not available in given mapping.")
        except AttributeError as e:
            raise KeyError(f"'{self._name}' has no attribute {_special_mapping[key]}") from e
        except Exception as e:
            raise KeyError(f"Failed to lookup '{key}' in '{self._name}'") from e

    def __getattr__(self, name: str) -> t.Any:
        if name in _reserved_namespace:
            raise ForbiddenAttributeError(
                f"'{name}' is a reserved namespace for {self._name} and should not be access nor modified."
            )
        if name in self._objects:
            return self._objects.__getitem__(name)
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._class_to_module.keys():
            module = self._get_module(self._class_to_module.__getitem__(name))
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)
        return value

    def _get_module(self, module_name: str):
        try:
            return importlib.import_module("." + module_name, self.__name__)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import {self.__name__}.{module_name} because of the following error (look up to see its"
                f" traceback):\n{e}"
            ) from e

    def __reduce__(self):
        return (self.__class__, (self._name, self.__file__, self._import_structure))



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
import logging
import types
import typing as t
from collections import OrderedDict

import inflection

import onediffusion

from .configuration_auto import AutoConfig


if t.TYPE_CHECKING:
    from collections import _odict_items
    from collections import _odict_keys
    from collections import _odict_values

    ConfigModelOrderedDict = OrderedDict[type[onediffusion.SDConfig], type[onediffusion.SD[t.Any, t.Any]]]
    ConfigModelKeysView = _odict_keys[type[onediffusion.SDConfig], type[onediffusion.SD[t.Any, t.Any]]]
    ConfigModelValuesView = _odict_values[type[onediffusion.SDConfig], type[onediffusion.SD[t.Any, t.Any]]]
    ConfigModelItemsView = _odict_items[type[onediffusion.SDConfig], type[onediffusion.SD[t.Any, t.Any]]]
else:
    ConfigModelKeysView = ConfigModelValuesView = ConfigModelItemsView = t.Any
    ConfigModelOrderedDict = OrderedDict

logger = logging.getLogger(__name__)


class _BaseAutoSDClass:
    _model_mapping: _LazyAutoMapping

    def __init__(self, *args: t.Any, **attrs: t.Any):
        raise EnvironmentError(
            f"Cannot instantiate {self.__class__.__name__} directly. "
            "Please use '{self.__class__.__name__}.Runner(model_name)' instead."
        )

    @t.overload
    @classmethod
    def for_model(
        cls,
        model_name: str,
        model_id: str | None = None,
        pipeline: str | None = None,
        return_runner_kwargs: t.Literal[False] = ...,
        sd_config: onediffusion.SDConfig | None = ...,
        ensure_available: t.Literal[False, True] = ...,
        **attrs: t.Any,
    ) -> onediffusion.SD[t.Any, t.Any]:
        ...

    @t.overload
    @classmethod
    def for_model(
        cls,
        model_name: str,
        model_id: str | None = None,
        pipeline: str | None = None,
        return_runner_kwargs: t.Literal[True] = ...,
        sd_config: onediffusion.SDConfig | None = ...,
        ensure_available: t.Literal[False, True] = ...,
        **attrs: t.Any,
    ) -> tuple[onediffusion.SD[t.Any, t.Any], dict[str, t.Any]]:
        ...

    @classmethod
    def for_model(
        cls,
        model_name: str,
        model_id: str | None = None,
        pipeline: str | None = None,
        return_runner_kwargs: bool = False,
        sd_config: onediffusion.SDConfig | None = None,
        ensure_available: bool = False,
        **attrs: t.Any,
    ) -> onediffusion.SD[t.Any, t.Any] | tuple[onediffusion.SD[t.Any, t.Any], dict[str, t.Any]]:
        """The lower level API for creating a OneDiffusion instance.

        ```python
        >>> import onediffusion
        >>> sd = onediffusion.AutoSD.for_model("stable-diffusion")
        ```
        """
        # order matters here
        runner_kwargs_name = {
            "models",
            "max_batch_size",
            "max_latency_ms",
            "method_configs",
            "scheduling_strategy",
        }
        to_runner_attrs = {k: v for k, v in attrs.items() if k in runner_kwargs_name}
        attrs = {k: v for k, v in attrs.items() if k not in to_runner_attrs}
        if cls._model_mapping.get(inflection.underscore(model_name), None, mapping_type="name2model"):
            if not isinstance(sd_config, onediffusion.SDConfig):
                # The rest of kwargs is now passed to config
                sd_config = AutoConfig.for_model(model_name, **attrs)
                attrs = sd_config.__sdserver_extras__
            # the rest of attrs will be saved to __sdserver_extras__
            sd = cls._model_mapping[type(sd_config)].from_pretrained(
                model_id,
                pipeline,
                sd_config=sd_config,
                **sd_config.__sdserver_extras__,
            )
            if ensure_available:
                logger.debug(
                    "'ensure_available=True', Downloading '%s' with 'model_id=%s' to local model store.",
                    model_name,
                    sd.model_id,
                )
                import bentoml
                from bentoml._internal.frameworks.diffusers_runners.utils import get_model_or_download
                _model_name = inflection.underscore(model_name)
                module = getattr(bentoml.diffusers_simple, _model_name)
                short_name = module.MODEL_SHORT_NAME
                get_model_or_download(short_name, sd.model_id)
            if not return_runner_kwargs:
                return sd
            return sd, to_runner_attrs
        raise ValueError(
            f"Unrecognized configuration class {sd_config.__class__} for this kind of AutoSD: {cls.__name__}.\n"
            f"SD type should be one of {', '.join(c.__name__ for c in cls._model_mapping.keys())}."
        )


    @classmethod
    def register(cls, config_class: type[onediffusion.SDConfig], sd_class: type[onediffusion.SD[t.Any, t.Any]]):
        """
        Register a new model for this class.

        Args:
            config_class: The configuration corresponding to the model to register.
            sd_class: The runnable to register.
        """
        if hasattr(sd_class, "config_class") and sd_class.config_class is not config_class:
            raise ValueError(
                "The model class you are passing has a `config_class` attribute that is not consistent with the "
                f"config class you passed (model has {sd_class.config_class} and you passed {config_class}. Fix "
                "one of those so they match!"
            )
        cls._model_mapping.register(config_class, sd_class)


def getattribute_from_module(module: types.ModuleType, attr: t.Any) -> t.Any:
    if attr is None:
        return None
    if isinstance(attr, tuple):
        return tuple(getattribute_from_module(module, a) for a in attr)
    if hasattr(module, attr):
        return getattr(module, attr)
    # Some of the mappings have entries model_type -> object of another model type. In that case we try to grab the
    # object at the top level.
    od_module = importlib.import_module("onediffusion")

    if module != od_module:
        try:
            return getattribute_from_module(od_module, attr)
        except ValueError:
            raise ValueError(f"Could not find {attr} neither in {module} nor in {od_module}!")
    else:
        raise ValueError(f"Could not find {attr} in {od_module}!")


class _LazyAutoMapping(ConfigModelOrderedDict):
    """Based on transformers.models.auto.configuration_auto._LazyAutoMapping
    This OrderedDict values() and keys() returns the list instead, so you don't
    have to do list(mapping.values()) to get the list of values.
    """

    def __init__(self, config_mapping: OrderedDict[str, str], model_mapping: OrderedDict[str, str]):
        self._config_mapping = config_mapping
        self._reverse_config_mapping = {v: k for k, v in config_mapping.items()}
        self._model_mapping = model_mapping
        self._extra_content: dict[t.Any, t.Any] = {}
        self._modules: dict[str, types.ModuleType] = {}

    def __len__(self):
        common_keys = set(self._config_mapping.keys()).intersection(self._model_mapping.keys())
        return len(common_keys) + len(self._extra_content)

    def __getitem__(self, key: type[onediffusion.SDConfig]) -> type[onediffusion.SD[t.Any, t.Any]]:
        if key in self._extra_content:
            return self._extra_content[key]
        model_type = self._reverse_config_mapping[key.__name__]
        if model_type in self._model_mapping:
            model_name = self._model_mapping[model_type]
            return self._load_attr_from_module(model_type, model_name)

        # Maybe there was several model types associated with this config.
        model_types = [k for k, v in self._config_mapping.items() if v == key.__name__]
        for mtype in model_types:
            if mtype in self._model_mapping:
                model_name = self._model_mapping[mtype]
                return self._load_attr_from_module(mtype, model_name)
        raise KeyError(key)

    def _load_attr_from_module(self, model_type: str, attr: str) -> t.Any:
        module_name = inflection.underscore(model_type)
        if module_name not in self._modules:
            self._modules[module_name] = importlib.import_module(f".{module_name}", "onediffusion.models")
        return getattribute_from_module(self._modules[module_name], attr)

    def keys(self):
        mapping_keys = [
            self._load_attr_from_module(key, name)
            for key, name in self._config_mapping.items()
            if key in self._model_mapping.keys()
        ]
        return t.cast(ConfigModelKeysView, mapping_keys + list(self._extra_content.keys()))

    @t.overload
    def get(
        self, key: type[onediffusion.SDConfig], default: t.Any, mapping_type: t.Literal["default"] = "default"
    ) -> type[onediffusion.SD[t.Any, t.Any]]:
        ...

    @t.overload
    def get(self, key: str, default: t.Any, mapping_type: t.Literal["name2model", "name2config"] = ...) -> str:
        ...

    def get(
        self,
        key: str | type[onediffusion.SDConfig],
        default: t.Any,
        mapping_type: t.Literal["default", "name2config", "name2model"] = "default",
    ) -> str | type[onediffusion.SD[t.Any, t.Any]]:
        _supported = {"default", "name2model", "name2config"}
        if mapping_type not in _supported:
            raise RuntimeError(f"Unknown mapping type {mapping_type} (supported: {_supported})")

        if mapping_type == "default":
            if t.TYPE_CHECKING:
                # we check for lenient_issubclass below, but pyright is too dumb to understand
                assert not isinstance(key, str)
            else:
                if not onediffusion.utils.lenient_issubclass(key, onediffusion.SDConfig):
                    raise KeyError(f"Key must be a type of 'onediffusion.SDConfig', got {key} instead.")
            try:
                return self.__getitem__(key)
            except KeyError:
                return default
        else:
            mapping = self._model_mapping if mapping_type == "name2model" else self._config_mapping
            assert isinstance(key, str), f"Key must be a string type if mapping_type={mapping_type}"
            try:
                return mapping.__getitem__(key)
            except KeyError:
                return default

    def __bool__(self):
        return bool(self.keys())

    def values(self):
        mapping_values = [
            self._load_attr_from_module(key, name)
            for key, name in self._model_mapping.items()
            if key in self._config_mapping.keys()
        ]
        return t.cast(ConfigModelValuesView, mapping_values + list(self._extra_content.values()))

    def items(self):
        mapping_items = [
            (
                self._load_attr_from_module(key, self._config_mapping[key]),
                self._load_attr_from_module(key, self._model_mapping[key]),
            )
            for key in self._model_mapping.keys()
            if key in self._config_mapping.keys()
        ]
        return t.cast(ConfigModelItemsView, mapping_items + list(self._extra_content.items()))

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, item: t.Any):
        if item in self._extra_content:
            return True
        if not hasattr(item, "__name__") or item.__name__ not in self._reverse_config_mapping:
            return False
        model_type = self._reverse_config_mapping[item.__name__]
        return model_type in self._model_mapping

    def register(self, key: t.Any, value: t.Any):
        """
        Register a new model in this mapping.
        """
        if hasattr(key, "__name__") and key.__name__ in self._reverse_config_mapping:
            model_type = self._reverse_config_mapping[key.__name__]
            if model_type in self._model_mapping.keys():
                raise ValueError(f"'{key}' is already used by a OneDiffusion model.")

        self._extra_content[key] = value


__all__ = ["_BaseAutoSDClass", "_LazyAutoMapping"]



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is derived from HuggingFace's AutoConfig, AutoModel, etc."""

from __future__ import annotations

import typing as t

import onediffusion

from ... import utils


_import_structure = {
    "configuration_auto": ["AutoConfig", "CONFIG_MAPPING", "CONFIG_MAPPING_NAMES"],
    "modeling_auto": ["MODEL_MAPPING_NAMES"],
}

try:
    if not utils.is_torch_available():
        raise onediffusion.exceptions.MissingDependencyError
except onediffusion.exceptions.MissingDependencyError:
    pass
else:
    _import_structure["modeling_auto"].extend(["AutoSD", "MODEL_MAPPING"])


if t.TYPE_CHECKING:
    from .configuration_auto import CONFIG_MAPPING as CONFIG_MAPPING
    from .configuration_auto import CONFIG_MAPPING_NAMES as CONFIG_MAPPING_NAMES
    from .configuration_auto import AutoConfig as AutoConfig
    from .modeling_auto import MODEL_MAPPING_NAMES as MODEL_MAPPING_NAMES
    from .modeling_flax_auto import MODEL_FLAX_MAPPING_NAMES as MODEL_FLAX_MAPPING_NAMES
    from .modeling_tf_auto import MODEL_TF_MAPPING_NAMES as MODEL_TF_MAPPING_NAMES

    try:
        if not utils.is_torch_available():
            raise onediffusion.exceptions.MissingDependencyError
    except onediffusion.exceptions.MissingDependencyError:
        pass
    else:
        from .modeling_auto import MODEL_MAPPING as MODEL_MAPPING
        from .modeling_auto import AutoSD as AutoSD

else:
    import sys

    sys.modules[__name__] = utils.LazyModule(__name__, globals()["__file__"], _import_structure, module_spec=__spec__)



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base exceptions for OneDiffusion. This extends BentoML exceptions.
"""
from __future__ import annotations

import bentoml


class OneDiffusionException(bentoml.exceptions.BentoMLException):
    """Base class for all OneDiffusion exceptions. This extends BentoMLException."""


class GpuNotAvailableError(OneDiffusionException):
    """Raised when there is no GPU available in given system."""


class ValidationError(OneDiffusionException):
    """Raised when a validation fails."""


class ForbiddenAttributeError(OneDiffusionException):
    """Raised when using an _internal field."""


class MissingAnnotationAttributeError(OneDiffusionException):
    """Raised when a field under onediffusion.SDConfig is missing annotations."""


class MissingDependencyError(BaseException):
    """Raised when a dependency is missing."""



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import importlib
import logging
import typing as t

import bentoml
import onediffusion


if t.TYPE_CHECKING:
    import torch
else:
    torch = onediffusion.utils.LazyLoader("torch", globals(), "torch")

logger = logging.getLogger(__name__)


class StableDiffusionXL(onediffusion.SD[t.Any, t.Any]):
    __sdserver_internal__ = True

    @property
    def import_kwargs(self):
        model_kwds = {
            "device_map": "auto",
            "torch_dtype": torch.float16,
        }
        return model_kwds

    def sd_post_init(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import types
import typing as t
from collections import OrderedDict

import inflection

import onediffusion


if t.TYPE_CHECKING:
    from collections import _odict_items
    from collections import _odict_keys
    from collections import _odict_values

    ConfigOrderedDict = OrderedDict[str, type[onediffusion.SDConfig]]

    ConfigKeysView = _odict_keys[str, type[onediffusion.SDConfig]]
    ConfigValuesView = _odict_values[str, type[onediffusion.SDConfig]]
    ConfigItemsView = _odict_items[str, type[onediffusion.SDConfig]]
else:
    ConfigKeysView = ConfigValuesView = ConfigItemsView = t.Any
    ConfigOrderedDict = OrderedDict

# NOTE: This is the entrypoint when adding new model config
CONFIG_MAPPING_NAMES = OrderedDict(
    [
        ("stable_diffusion", "StableDiffusionConfig"),
        ("stable_diffusion_xl", "StableDiffusionXLConfig"),
    ]
)


class _LazyConfigMapping(ConfigOrderedDict):
    def __init__(self, mapping: OrderedDict[str, str]):
        self._mapping = mapping
        self._extra_content: dict[str, t.Any] = {}
        self._modules: dict[str, types.ModuleType] = {}

    def __getitem__(self, key: str):
        if key in self._extra_content:
            return self._extra_content[key]
        if key not in self._mapping:
            raise KeyError(key)
        value = self._mapping[key]
        module_name = inflection.underscore(key)
        if module_name not in self._modules:
            self._modules[module_name] = onediffusion.utils.ModelEnv(module_name).module
        if hasattr(self._modules[module_name], value):
            return getattr(self._modules[module_name], value)

        # Some of the mappings have entries model_type -> config of another model type. In that case we try to grab the
        # object at the top level.
        return getattr(onediffusion, value)

    def keys(self):
        return t.cast(ConfigKeysView, list(self._mapping.keys()) + list(self._extra_content.keys()))

    def values(self):
        return t.cast(ConfigValuesView, [self[k] for k in self._mapping.keys()] + list(self._extra_content.values()))

    def items(self):
        return t.cast(ConfigItemsView, [(k, self[k]) for k in self._mapping.keys()] + list(self._extra_content.items()))

    def __iter__(self):
        return iter(list(self._mapping.keys()) + list(self._extra_content.keys()))

    def __contains__(self, item: t.Any):
        return item in self._mapping or item in self._extra_content

    def register(self, key: str, value: t.Any):
        """
        Register a new configuration in this mapping.
        """
        if key in self._mapping.keys():
            raise ValueError(f"'{key}' is already used by a OneDiffusion config, pick another name.")
        self._extra_content[key] = value


CONFIG_MAPPING: dict[str, type[onediffusion.SDConfig]] = _LazyConfigMapping(CONFIG_MAPPING_NAMES)

# The below handle special alias when we call underscore to the name directly
# without processing camelcase first.
CONFIG_NAME_ALIASES: dict[str, str] = {
    #"stable_diffusion": "stablediffusion",
    #"stable_diffusion_xl": "stablediffusionxl",
}


class AutoConfig:
    def __init__(self, *_: t.Any, **__: t.Any):
        raise EnvironmentError("Cannot instantiate Config. Please use `Config.for_model(model_name)` instead.")

    @classmethod
    def for_model(cls, model_name: str, **attrs: t.Any) -> onediffusion.SDConfig:
        model_name = inflection.underscore(model_name)
        if model_name in CONFIG_MAPPING:
            return CONFIG_MAPPING[model_name].model_construct_env(**attrs)
        raise ValueError(
            f"Unrecognized configuration class for {model_name}. "
            f"Model name should be one of {', '.join(CONFIG_MAPPING.keys())}."
        )

    @classmethod
    def infer_class_from_name(cls, name: str) -> type[onediffusion.SDConfig]:
        model_name = inflection.underscore(name)
        if model_name in CONFIG_NAME_ALIASES:
            model_name = CONFIG_NAME_ALIASES[model_name]

        if model_name in CONFIG_MAPPING:
            return CONFIG_MAPPING[model_name]
        raise ValueError(
            f"Unrecognized configuration class for {model_name}. "
            f"Model name should be one of {', '.join(CONFIG_MAPPING.keys())}."
        )



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Telemetry related for OneDiffusion tracking.

Users can disable this with ONEDIFFUSION_DO_NOT_TRACK envvar.
"""
from __future__ import annotations

import contextlib
import functools
import os
import typing as t

import attr

from bentoml._internal.utils import analytics as _internal_analytics
from bentoml._internal.utils.analytics import usage_stats as _internal_usage


if t.TYPE_CHECKING:
    import onediffusion

from ..__about__ import __version__


ENV_VARS_TRUE_VALUES = {"1", "ON", "YES", "TRUE"}

# This variable is a proxy that will control BENTOML_DO_NOT_TRACK
ONEDIFFUSION_DO_NOT_TRACK = "ONEDIFFUSION_DO_NOT_TRACK"

DO_NOT_TRACK = os.environ.get(ONEDIFFUSION_DO_NOT_TRACK, str(False)).upper()


@functools.lru_cache(maxsize=1)
def do_not_track() -> bool:
    return DO_NOT_TRACK in ENV_VARS_TRUE_VALUES


@_internal_usage.silent
def track(event_properties: _internal_analytics.schemas.EventMeta):
    if do_not_track():
        return
    _internal_analytics.track(event_properties)


@contextlib.contextmanager
def set_bentoml_tracking():
    original_value = os.environ.pop(_internal_analytics.BENTOML_DO_NOT_TRACK, str(False))
    try:
        os.environ[_internal_analytics.BENTOML_DO_NOT_TRACK] = str(do_not_track())
        yield
    finally:
        os.environ[_internal_analytics.BENTOML_DO_NOT_TRACK] = original_value


@attr.define
class OneDiffusionCliEvent(_internal_analytics.schemas.EventMeta):
    cmd_group: str
    cmd_name: str
    onediffusion_version: str = __version__

    # NOTE: reserved for the do_not_track logics
    duration_in_ms: t.Any = attr.field(default=None)
    error_type: str = attr.field(default=None)
    return_code: int = attr.field(default=None)


@attr.define
class StartInitEvent(_internal_analytics.schemas.EventMeta):
    model_name: str
    sd_config: t.Dict[str, t.Any] = attr.field(default=None)

    @staticmethod
    def handler(sd_config: onediffusion.SDConfig) -> StartInitEvent:
        return StartInitEvent(model_name=sd_config["model_name"], sd_config=sd_config.model_dump())


def track_start_init(
    sd_config: onediffusion.SDConfig,
):
    if do_not_track():
        return
    track(StartInitEvent.handler(sd_config))



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The following includes OneDiffusion configuration and excerpt from
[Stable Diffusion XL repository](https://github.com/Stability-AI/generative-models)
"""
from __future__ import annotations

import onediffusion


class StableDiffusionXLConfig(onediffusion.SDConfig):
    """Stable Diffusion XL is a latent text-to-image diffusion model capable of generating photo-realistic images given any text input.

    Refer to [Stable Diffusion XL page](https://huggingface.co/stabilityai/stable-diffusion-xl-base-1.0) for more information.
    """

    __config__ = {
        "timeout": 3600000,
        "url": "https://github.com/Stability-AI/generative-models",
        "default_id": "stabilityai/stable-diffusion-xl-base-1.0",
        "model_ids": ["stabilityai/stable-diffusion-xl-base-1.0"],
        "default_pipeline": "text2img",
        "pipelines": ["text2img", "img2img"],
    }


START_STABLE_DIFFUSION_XL_COMMAND_DOCSTRING = """\
Run a OneDiffusion server for Stable Diffusion XL model.

\b
> See more information about Stable Diffusion XL at [stabilityai/stable-diffusion-xl-base-1.0](https://huggingface.co/stabilityai/stable-diffusion-xl-base-1.0)

\b
## Usage

Currently, stable-diffusion-xl only supports PyTorch. Make sure ``torch`` is available in your system.

\b
stable-diffusion-xl will use stabilityai/stable-diffusion-xl-base-1.0 as the default model. To change any to any other Stable Diffusion XL
saved pretrained, or a fine-tune model, provide ``ONEDIFFUSION_STABLE_DIFFUSION_XL_MODEL_ID='stabilityai/stable-diffusion-xl-base-1.0'``
or provide `--model-id` flag when running ``onediffusion start stable-diffusion-xl``:

\b
$ onediffusion start stable-diffusion-xl --model-id stabilityai/stable-diffusion-xl-base-1.0
"""



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Configuration utilities for SDConfig. All model configuration will inherit from ``onediffusion.SDConfig``.

Highlight feature: Each fields in ``onediffusion.SDConfig`` will also automatically generate a environment
variable based on its name field.

For example, the following config class:

```python
class StableDiffusionConfig(onediffusion.SDConfig):
    __config__ = {
        "timeout": 3600000,
        "url": "https://github.com/Stability-AI/stablediffusion",
        "default_id": "stabilityai/stable-diffusion-2",
        "model_ids": [
            "CompVis/stable-diffusion-v1-4",
            "runwayml/stable-diffusion-v1-5",
            "stabilityai/stable-diffusion-2",
        ],
        "default_pipeline": "text2img",
        "pipelines": ["text2img", "img2img"],
    }
```

which generates the environment ONEDIFFUSION_FLAN_T5_GENERATION_TEMPERATURE for users to configure temperature
dynamically during serve, ahead-of-serve or per requests.

Refer to ``onediffusion.SDConfig`` docstring for more information.
"""
from __future__ import annotations

import functools
import logging
import os
import sys
import typing as t
from operator import itemgetter

import attr
import click_option_group as cog
import inflection
import orjson
from cattr.gen import make_dict_unstructure_fn
from cattr.gen import override
from deepmerge.merger import Merger

import onediffusion

from .exceptions import ForbiddenAttributeError
from .utils import DEBUG
from .utils import ENV_VARS_TRUE_VALUES
from .utils import LazyType
from .utils import bentoml_cattr
from .utils import codegen
from .utils import dantic
from .utils import first_not_none
from .utils import lenient_issubclass
from .utils import non_intrusive_setattr


if hasattr(t, "Required"):
    from typing import Required
else:
    from typing_extensions import Required

if hasattr(t, "NotRequired"):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired

if hasattr(t, "dataclass_transform"):
    from typing import dataclass_transform
else:
    from typing_extensions import dataclass_transform

_T = t.TypeVar("_T")


if t.TYPE_CHECKING:
    from attr import _CountingAttr  # type: ignore
    from attr import _make_init  # type: ignore
    from attr import _make_repr  # type: ignore
    from attr import _transform_attrs  # type: ignore

    from ._types import ClickFunctionWrapper
    from ._types import F
    from ._types import O_co
    from ._types import P

    ReprArgs: t.TypeAlias = t.Iterable[tuple[str | None, t.Any]]

    DictStrAny = dict[str, t.Any]
    ListStr = list[str]
    ItemgetterAny = itemgetter[t.Any]
    FieldTransformers = t.Callable[[_T, list[attr.Attribute[t.Any]]], list[attr.Attribute[t.Any]]]
else:
    Constraint = t.Any
    ListStr = list
    DictStrAny = dict
    ItemgetterAny = itemgetter
    # NOTE: Using internal API from attr here, since we are actually
    # allowing subclass of onediffusion.SDConfig to become 'attrs'-ish
    from attr._make import _CountingAttr
    from attr._make import _make_init
    from attr._make import _make_repr
    from attr._make import _transform_attrs


__all__ = ["SDConfig"]

logger = logging.getLogger(__name__)

config_merger = Merger(
    # merge dicts
    type_strategies=[(DictStrAny, "merge")],
    # override all other types
    fallback_strategies=["override"],
    # override conflicting types
    type_conflict_strategies=["override"],
)


def _field_env_key(model_name: str, key: str, suffix: str | t.Literal[""] | None = None) -> str:
    return "_".join(filter(None, map(str.upper, ["ONEDIFFUSION", model_name, suffix.strip("_") if suffix else "", key])))


# cached it here to save one lookup per assignment
_object_getattribute = object.__getattribute__


class ModelSettings(t.TypedDict, total=False):
    """ModelSettings serve only for typing purposes as this is transcribed into SDConfig.__config__.
    Note that all fields from this dictionary will then be converted to __sdserver_*__ fields in SDConfig.
    """

    # NOTE: These required fields should be at the top, as it will be kw_only
    default_id: Required[str]
    model_ids: Required[ListStr]
    default_pipeline: Required[str]
    pipelines: Required[ListStr]

    # meta
    url: str
    requires_gpu: bool
    service_name: NotRequired[str]
    requirements: t.Optional[ListStr]

    # naming convention, only name_type is needed to infer from the class
    # as the three below it can be determined automatically
    name_type: t.Literal["dasherize", "lowercase"]
    model_name: NotRequired[str]
    start_name: NotRequired[str]
    env: NotRequired[onediffusion.utils.ModelEnv]

    # serving configuration
    timeout: int
    workers_per_resource: t.Union[int, float]


def _settings_field_transformer(
    _: type[attr.AttrsInstance], __: list[attr.Attribute[t.Any]]
) -> list[attr.Attribute[t.Any]]:
    return [
        attr.Attribute.from_counting_attr(
            k,
            dantic.Field(
                kw_only=False if t.get_origin(ann) is not Required else True,
                auto_default=True,
                use_default_converter=False,
                type=ann,
                metadata={"target": f"__sdserver_{k}__"},
                description=f"ModelSettings field for {k}.",
            ),
        )
        for k, ann in t.get_type_hints(ModelSettings).items()
    ]


@attr.define(slots=True, field_transformer=_settings_field_transformer, frozen=False)
class _ModelSettingsAttr:
    """Internal attrs representation of ModelSettings."""

    def __getitem__(self, key: str) -> t.Any:
        if key in codegen.get_annotations(ModelSettings):
            return _object_getattribute(self, key)
        raise KeyError(key)

    @classmethod
    def default(cls) -> _ModelSettingsAttr:
        _ = ModelSettings(
            default_id="__default__",
            model_ids=["__default__"],
            default_pipeline="__default__",
            pipelines=["__default__"],
            name_type="dasherize",
            requires_gpu=False,
            url="",
            requirements=None,
            timeout=int(36e6),
            service_name="",
            workers_per_resource=1,
        )
        return cls(**t.cast(DictStrAny, _))


def structure_settings(cl_: type[SDConfig], cls: type[_ModelSettingsAttr]):
    if not lenient_issubclass(cl_, SDConfig):
        raise RuntimeError(f"Given '{cl_}' must be a subclass type of 'SDConfig', got '{cl_}' instead.")

    if not hasattr(cl_, "__config__") or getattr(cl_, "__config__") is None:
        raise RuntimeError("Given SDConfig must have '__config__' that is not None defined.")

    assert cl_.__config__ is not None

    _cl_name = cl_.__name__.replace("Config", "")

    _settings_attr = _ModelSettingsAttr.default()
    try:
        cls(**t.cast(DictStrAny, cl_.__config__))
        _settings_attr = attr.evolve(_settings_attr, **t.cast(DictStrAny, cl_.__config__))
    except TypeError as e:
        raise ValueError("Either 'default_id' or 'model_ids' are emptied under '__config__' (required fields).")

    _final_value_dct: DictStrAny = {
        "model_name": inflection.underscore(_cl_name)
        if _settings_attr["name_type"] == "dasherize"
        else _cl_name.lower()
    }
    _final_value_dct["start_name"] = (
        inflection.dasherize(_final_value_dct["model_name"])
        if _settings_attr["name_type"] == "dasherize"
        else _final_value_dct["model_name"]
    )
    env = onediffusion.utils.ModelEnv(_final_value_dct["model_name"])
    _final_value_dct["env"] = env

    _final_value_dct["service_name"] = f"generated_{_final_value_dct['model_name']}_service.py"

    return attr.evolve(_settings_attr, **_final_value_dct)


bentoml_cattr.register_structure_hook(_ModelSettingsAttr, structure_settings)


def _make_env_transformer(
    cls: type[SDConfig],
    model_name: str,
    suffix: t.LiteralString | None = None,
    default_callback: t.Callable[[str, t.Any], t.Any] | None = None,
    globs: DictStrAny | None = None,
):
    def identity(_: str, x_value: t.Any) -> t.Any:
        return x_value

    default_callback = identity if default_callback is None else default_callback

    globs = {} if globs is None else globs
    globs.update(
        {
            "functools": functools,
            "__populate_env": dantic.env_converter,
            "__default_callback": default_callback,
            "__field_env": _field_env_key,
            "__suffix": suffix or "",
            "__model_name": model_name,
        }
    )

    lines: ListStr = [
        "__env = lambda field_name: __field_env(__model_name, field_name, __suffix)",
        "return [",
        "    f.evolve(",
        "        default=__populate_env(__default_callback(f.name, f.default), __env(f.name)),",
        "        metadata={",
        "            'env': f.metadata.get('env', __env(f.name)),",
        "            'description': f.metadata.get('description', '(not provided)'),",
        "        },",
        "    )",
        "    for f in fields",
        "]",
    ]
    fields_ann = "list[attr.Attribute[t.Any]]"

    return codegen.generate_function(
        cls,
        "__auto_env",
        lines,
        args=("_", "fields"),
        globs=globs,
        annotations={"_": "type[SDConfig]", "fields": fields_ann, "return": fields_ann},
    )


def _setattr_class(attr_name: str, value_var: t.Any, add_dunder: bool = False):
    """
    Use the builtin setattr to set *attr_name* to *value_var*.
    We can't use the cached object.__setattr__ since we are setting
    attributes to a class.

    If add_dunder to True, the generated globs should include a __add_dunder
    value that will be used to add the dunder methods to the class for given
    value_var
    """
    val = f"__add_dunder(cls, {value_var})" if add_dunder else value_var
    return f"setattr(cls, '{attr_name}', {val})"


_dunder_add = {}


def _make_assignment_script(
    cls: type[SDConfig], attributes: attr.AttrsInstance, _prefix: t.LiteralString = "sdserver"
) -> t.Callable[..., None]:
    """Generate the assignment script with prefix attributes __sdserver_<value>__"""
    args: ListStr = []
    globs: DictStrAny = {
        "cls": cls,
        "_cached_attribute": attributes,
        "_cached_getattribute_get": _object_getattribute.__get__,
        "__add_dunder": codegen.add_method_dunders,
    }
    annotations: DictStrAny = {"return": None}

    lines: ListStr = ["_getattr = _cached_getattribute_get(_cached_attribute)"]
    for attr_name, field in attr.fields_dict(attributes.__class__).items():
        arg_name = field.metadata.get("target", f"__{_prefix}_{inflection.underscore(attr_name)}__")
        args.append(f"{attr_name}=getattr(_cached_attribute, '{attr_name}')")
        lines.append(_setattr_class(arg_name, attr_name, add_dunder=attr_name in _dunder_add))
        annotations[attr_name] = field.type

    return codegen.generate_function(
        cls, "__assign_attr", lines, args=("cls", *args), globs=globs, annotations=annotations
    )


_reserved_namespace = {"__config__"}


@dataclass_transform(order_default=True, field_specifiers=(attr.field, dantic.Field))
def __sd_config_transform__(cls: type[SDConfig]) -> type[SDConfig]:
    kwargs: dict[str, t.Any] = {}
    non_intrusive_setattr(
        cls,
        "__dataclass_transform__",
        {
            "order_default": True,
            "field_specifiers": (attr.field, dantic.Field),
            "kwargs": kwargs,
        },
    )
    return cls


@attr.define(slots=True)
class SDConfig:
    """
    ``onediffusion.SDConfig`` is somewhat a hybrid combination between the performance of `attrs` with the
    easy-to-use interface that pydantic offer. It lives in between where it allows users to quickly formulate
    a SDConfig for any diffusion model without worrying too much about performance. It does a few things:

    - Automatic environment conversion: Each fields will automatically be provisioned with an environment
        variable, make it easy to work with ahead-of-time or during serving time
    - Familiar API: It is compatible with cattrs as well as providing a few Pydantic-2 like API,
        i.e: ``model_construct_env``, ``to_generation_config``, ``to_click_options``
    - Automatic CLI generation: It can identify each fields and convert it to compatible Click options.
        This means developers can use any of the SDConfig to create CLI with compatible-Python
        CLI library (click, typer, ...)

    > Internally, SDConfig is an attrs class. All subclass of SDConfig contains "attrs-like" features,
    > which means SDConfig will actually generate subclass to have attrs-compatible API, so that the subclass
    > can be written as any normal Python class.

    By default, all SDConfig must provide a __config__ with 'default_id' and 'model_ids'.

    All other fields are optional, and will be use default value if not set.

    ```python
    class StableDiffusionConfig(onediffusion.SDConfig):

        __config__ = {
            "timeout": 3600000,
            "url": "https://github.com/Stability-AI/stablediffusion",
            "default_id": "stabilityai/stable-diffusion-2",
            "model_ids": [
                "CompVis/stable-diffusion-v1-4",
                "runwayml/stable-diffusion-v1-5",
                "stabilityai/stable-diffusion-2",
            ],
            "default_pipeline": "text2img",
            "pipelines": ["text2img", "img2img"],
        }
    ```
    """

    Field = dantic.Field
    """Field is a alias to the internal dantic utilities to easily create
    attrs.fields with pydantic-compatible interface. For example:

    ```python
    class MyModelConfig(onediffusion.SDConfig):

        field1 = onediffusion.SDConfig.Field(...)
    ```
    """

    # NOTE: The following is handled via __init_subclass__, and is only used for TYPE_CHECKING
    if t.TYPE_CHECKING:
        # NOTE: public attributes to override
        __config__: ModelSettings | None = Field(None)
        """Internal configuration for this diffusion model. Each of the field in here will be populated
        and prefixed with __sdserver_<value>__"""

        """Users can override this subclass of any given SDConfig to provide GenerationConfig
        default value. For example:

        ```python
        class MyAwesomeModelConfig(onediffusion.SDConfig):
            class GenerationConfig:
                max_new_tokens: int = 200
                top_k: int = 10
                num_return_sequences: int = 1
                eos_token_id: int = 11
        ```
        """

        # NOTE: Internal attributes that should only be used by Onediffusion. Users usually shouldn't
        # concern any of these. These are here for pyright not to complain.
        def __attrs_init__(self, **attrs: t.Any):
            """Generated __attrs_init__ for SDConfig subclass that follows the attrs contract."""

        __attrs_attrs__: tuple[attr.Attribute[t.Any], ...] = Field(None, init=False)
        """Since we are writing our own __init_subclass__, which is an alternative way for __prepare__,
        we want onediffusion.SDConfig to be attrs-like dataclass that has pydantic-like interface.
        __attrs_attrs__ will be handled dynamically by __init_subclass__.
        """

        __sdserver_hints__: DictStrAny = Field(None, init=False)
        """An internal cache of resolved types for this SDConfig."""

        __sdserver_accepted_keys__: set[str] = Field(None, init=False)
        """The accepted keys for this SDConfig."""

        __sdserver_extras__: DictStrAny = Field(None, init=False)
        """Extra metadata for this SDConfig."""

        # NOTE: The following will be populated from __config__ and also
        # considered to be public API.
        __sdserver_default_id__: str = Field(None)
        """Return the default model to use when using 'onediffusion start <model_id>'.
        This could be one of the keys in 'self.model_ids' or custom users model.

        This field is required when defining under '__config__'.
        """

        __sdserver_model_ids__: ListStr = Field(None)
        """A list of supported pretrained models tag for this given runnable.

        For example:
            For FLAN-T5 impl, this would be ["google/flan-t5-small", "google/flan-t5-base",
                                                "google/flan-t5-large", "google/flan-t5-xl", "google/flan-t5-xxl"]

        This field is required when defining under '__config__'.
        """

        __sdserver_default_pipeline__: str = Field(None)
        """Return the default pipeline to use when using 'onediffusion start <model_id>'.

        This field is required when defining under '__config__'.
        """

        __sdserver_pipelines__: ListStr = Field(None)
        """A list of supported pipelines for this given runnable.
        """

        __sdserver_url__: str = Field(None, init=False)
        """The resolved url for this SDConfig."""

        __sdserver_requires_gpu__: bool = Field(None, init=False)
        """Determines if this model is only available on GPU. By default it supports GPU and fallback to CPU."""

        __sdserver_service_name__: str = Field(None)
        """Generated service name for this SDConfig. By default, it is 'generated_{model_name}_service.py'"""

        __sdserver_requirements__: ListStr | None = Field(None)
        """The default PyPI requirements needed to run this given SD. By default, we will depend on
        bentoml, torch, transformers, diffusers, safetensors, accelerate, triton, pydantic."""

        __sdserver_name_type__: t.Literal["dasherize", "lowercase"] = Field("dasherize")
        """the default name typed for this model. "dasherize" will convert the name to lowercase and
        replace spaces with dashes. "lowercase" will convert the name to lowercase."""

        __sdserver_model_name__: str = Field(None)
        """The normalized version of __sdserver_start_name__, determined by __sdserver_name_type__"""

        __sdserver_start_name__: str = Field(None)
        """Default name to be used with `onediffusion start`"""

        __sdserver_env__: onediffusion.utils.ModelEnv = Field(None, init=False)
        """A ModelEnv instance for this SDConfig."""

        __sdserver_timeout__: int = Field(36e6)
        """The default timeout to be set for this given diffusion model."""

        __sdserver_workers_per_resource__: int | float = Field(1)
        """The number of workers per resource. This is used to determine the number of workers to use for this model.
        For example, if this is set to 0.5, then Sdserver will use 1 worker per 2 resources. If this is set to 1, then
        Sdserver will use 1 worker per resource. If this is set to 2, then Sdserver will use 2 workers per resource.

        See https://docs.bentoml.org/en/latest/guides/scheduling.html#resource-scheduling-strategy for more details.

        By default, it is set to 1.
        """



    def __init_subclass__(cls):
        """The purpose of this __init_subclass__ is that we want all subclass of SDConfig
        to adhere to the attrs contract, and have pydantic-like interface. This means we will
        construct all fields and metadata and hack into how attrs use some of the 'magic' construction
        to generate the fields.

        It also does a few more extra features: It also generate all __sdserver_*__ config from
        ModelSettings (derived from __config__) to the class.
        """
        if not cls.__name__.endswith("Config"):
            logger.warning("SDConfig subclass should end with 'Config'. Updating to %sConfig", cls.__name__)
            cls.__name__ = f"{cls.__name__}Config"

        # NOTE: auto assignment attributes generated from __config__
        _make_assignment_script(cls, bentoml_cattr.structure(cls, _ModelSettingsAttr))(cls)
        # process a fields under cls.__dict__ and auto convert them with dantic.Field
        cd = cls.__dict__
        anns = codegen.get_annotations(cls)

        # _CountingAttr is the underlying representation of attr.field
        ca_names = {name for name, attr in cd.items() if isinstance(attr, _CountingAttr)}
        these: dict[str, _CountingAttr[t.Any]] = {}
        annotated_names: set[str] = set()
        for attr_name, typ in anns.items():
            if codegen.is_class_var(typ):
                continue
            annotated_names.add(attr_name)
            val = cd.get(attr_name, attr.NOTHING)
            if not LazyType["_CountingAttr[t.Any]"](_CountingAttr).isinstance(val):
                if val is attr.NOTHING:
                    val = cls.Field(env=_field_env_key(cls.__sdserver_model_name__, attr_name))
                else:
                    val = cls.Field(default=val, env=_field_env_key(cls.__sdserver_model_name__, attr_name))
            these[attr_name] = val
        unannotated = ca_names - annotated_names
        if len(unannotated) > 0:
            missing_annotated = sorted(unannotated, key=lambda n: t.cast("_CountingAttr[t.Any]", cd.get(n)).counter)
            raise onediffusion.exceptions.MissingAnnotationAttributeError(
                f"The following field doesn't have a type annotation: {missing_annotated}"
            )

        cls.__sdserver_accepted_keys__ = set(these.keys())

        # Generate the base __attrs_attrs__ transformation here.
        attrs, base_attrs, base_attr_map = _transform_attrs(
            cls,  # the current class
            these,  # the parsed attributes we previous did
            False,  # disable auto_attribs, since we already handle these
            False,  # disable kw_only
            True,  # collect_by_mro
            field_transformer=_make_env_transformer(cls, cls.__sdserver_model_name__),
        )
        _weakref_slot = True  # slots = True
        _base_names = {a.name for a in base_attrs}
        _attr_names = tuple(a.name for a in attrs)
        _has_pre_init = bool(getattr(cls, "__attrs_pre_init__", False))
        _has_post_init = bool(getattr(cls, "__attrs_post_init__", False))
        # the protocol for attrs-decorated class
        cls.__attrs_attrs__ = attrs
        # generate a __attrs_init__ for the subclass, since we will
        # implement a custom __init__
        cls.__attrs_init__ = codegen.add_method_dunders(
            cls,
            _make_init(
                cls,  # cls (the attrs-decorated class)
                attrs,  # tuple of attr.Attribute of cls
                _has_pre_init,  # pre_init
                _has_post_init,  # post_init
                False,  # frozen
                True,  # slots
                False,  # cache_hash
                base_attr_map,  # base_attr_map
                False,  # is_exc (check if it is exception)
                None,  # cls_on_setattr (essentially attr.setters)
                attrs_init=True,  # whether to create __attrs_init__ instead of __init__
            ),
        )
        # __repr__ function with the updated fields.
        cls.__repr__ = codegen.add_method_dunders(cls, _make_repr(cls.__attrs_attrs__, None, cls))
        # Traverse the MRO to collect existing slots
        # and check for an existing __weakref__.
        existing_slots: DictStrAny = {}
        weakref_inherited = False
        for base_cls in cls.__mro__[1:-1]:
            if base_cls.__dict__.get("__weakref__", None) is not None:
                weakref_inherited = True
            existing_slots.update({name: getattr(base_cls, name) for name in getattr(base_cls, "__slots__", [])})

        if (
            _weakref_slot
            and "__weakref__" not in getattr(cls, "__slots__", ())
            and "__weakref__" not in _attr_names
            and not weakref_inherited
        ):
            _attr_names += ("__weakref__",)

        # We only add the names of attributes that aren't inherited.
        # Setting __slots__ to inherited attributes wastes memory.
        slot_names = [name for name in _attr_names if name not in _base_names]
        # There are slots for attributes from current class
        # that are defined in parent classes.
        # As their descriptors may be overridden by a child class,
        # we collect them here and update the class dict
        reused_slots = {
            slot: slot_descriptor for slot, slot_descriptor in existing_slots.items() if slot in slot_names
        }
        # __sdserver_extras__ holds additional metadata that might be usefule for users, hence we add it to slots
        slot_names = [name for name in slot_names if name not in reused_slots] + ["__sdserver_extras__"]
        cls.__slots__ = tuple(slot_names)
        # Finally, resolve the types
        if getattr(cls, "__attrs_types_resolved__", None) != cls:
            # NOTE: We will try to resolve type here, and cached it for faster use
            # It will be about 15-20ms slower comparing not resolving types.
            globs: DictStrAny = {"t": t, "typing": t, "Constraint": Constraint}
            if cls.__module__ in sys.modules:
                globs.update(sys.modules[cls.__module__].__dict__)

            cls = attr.resolve_types(cls, globalns=globs)

        # the hint cache for easier access
        cls.__sdserver_hints__ = {
            f.name: f.type for ite in map(attr.fields, (cls, )) for f in ite
        }
        cls = __sd_config_transform__(cls)

    def __setattr__(self, attr: str, value: t.Any):
        if attr in _reserved_namespace:
            raise ForbiddenAttributeError(
                f"{attr} should not be set during runtime "
                f"as these value will be reflected during runtime. "
                f"Instead, you can create a custom SD subclass {self.__class__.__name__}."
            )

        super().__setattr__(attr, value)

    def __init__(
        self,
        *,
        __sdserver_extras__: DictStrAny | None = None,
        **attrs: t.Any,
    ):
        # create a copy of the keys as cache
        _cached_keys = tuple(attrs.keys())
        _cached_keys = tuple(k for k in _cached_keys if k in attrs)

        self.__sdserver_extras__ = first_not_none(__sdserver_extras__, default={})
        config_merger.merge(
            self.__sdserver_extras__, {k: v for k, v in attrs.items() if k not in self.__sdserver_accepted_keys__}
        )

        for k in _cached_keys:
            if k in self.__sdserver_extras__:
                del attrs[k]
        _cached_keys = tuple(k for k in _cached_keys if k in attrs)

        if DEBUG:
            logger.info(
                "Creating %s with the following attributes: %s",
                self.__class__.__name__,
                _cached_keys,
            )

        # The rest of attrs should only be the attributes to be passed to __attrs_init__
        self.__attrs_init__(**attrs)

    def __getitem__(self, item: str | t.Any) -> t.Any:
        """Allowing access SDConfig as a dictionary. The order will always evaluate as

        __sdserver_*__ > self.key > __sdserver_extras__

        This method is purely for convenience, and should not be used for performance critical code.
        """
        if not isinstance(item, str):
            raise TypeError(f"SD only supports string indexing, not {item.__class__.__name__}")
        if item in _reserved_namespace:
            raise ForbiddenAttributeError(
                f"'{item}' is a reserved namespace for {self.__class__} and should not be access nor modified."
            )
        internal_attributes = f"__sdserver_{item}__"
        if hasattr(self, internal_attributes):
            return getattr(self, internal_attributes)
        elif hasattr(self, item):
            return getattr(self, item)
        elif item in self.__sdserver_extras__:
            return self.__sdserver_extras__[item]
        else:
            raise KeyError(item)

    def __getattribute__(self, item: str) -> t.Any:
        if item in _reserved_namespace:
            raise ForbiddenAttributeError(
                f"'{item}' belongs to a private namespace for {self.__class__} and should not be access nor modified."
            )
        return _object_getattribute.__get__(self)(item)

    @classmethod
    def model_derivate(cls, name: str | None = None, **attrs: t.Any) -> SDConfig:
        """A helper class to generate a new SDConfig class with additional attributes.

        This is useful to modify builtin __config__ value attributes.

        ```python
        class DollyV2Config(onediffusion.SDConfig):
            ...

        my_new_class = DollyV2Config.model_derivate(default_id='...')
        ```

        Args:
            name: The name of the new class.
            **attrs: The attributes to be added to the new class. This will override
                     any existing attributes with the same name.
        """
        assert cls.__config__ is not None, "Cannot derivate a SDConfig without __config__"
        _new_cfg = {k: v for k, v in attrs.items() if k in attr.fields_dict(_ModelSettingsAttr)}
        attrs = {k: v for k, v in attrs.items() if k not in _new_cfg}
        new_cls = types.new_class(
            name or f"{cls.__name__.replace('Config', '')}DerivateConfig",
            (cls,),
            {},
            lambda ns: ns.update(
                {
                    "__config__": config_merger.merge(copy.deepcopy(cls.__dict__["__config__"]), _new_cfg),
                }
            ),
        )

        # For pickling to work, the __module__ variable needs to be set to the
        # frame where the class is created.  Bypass this step in environments where
        # sys._getframe is not defined (Jython for example) or sys._getframe is not
        # defined for arguments greater than 0 (IronPython).
        try:
            new_cls.__module__ = sys._getframe(1).f_globals.get("__name__", "__main__")
        except (AttributeError, ValueError):
            pass

        return new_cls(**attrs)

    def model_dump(self, flatten: bool = False, **_: t.Any):
        dumped = bentoml_cattr.unstructure(self)
        return dumped

    def model_dump_json(self, **kwargs: t.Any):
        return orjson.dumps(self.model_dump(**kwargs))

    @classmethod
    def model_construct_env(cls, **attrs: t.Any) -> t.Self:
        """A helpers that respect configuration values that
        sets from environment variables for any given configuration class.
        """
        attrs = {k: v for k, v in attrs.items() if v is not None}
        model_config = cls.__sdserver_env__.config

        env_json_string = os.environ.get(model_config, None)

        if env_json_string is not None:
            try:
                config_from_env = orjson.loads(env_json_string)
            except orjson.JSONDecodeError as e:
                raise RuntimeError(f"Failed to parse '{model_config}' as valid JSON string.") from e
        else:
            config_from_env = {}

        env_struct = bentoml_cattr.structure(config_from_env, cls)

        return attr.evolve(env_struct, **attrs)

    def model_validate_click(self, **attrs: t.Any) -> tuple[SDConfig, DictStrAny]:
        """Parse given click attributes into a SDConfig and return the remaining click attributes."""
        sd_config_attrs: DictStrAny = {}
        key_to_remove: ListStr = []

        for k, v in attrs.items():
            if k.startswith(f"{self['model_name']}_"):
                sd_config_attrs[k[len(self["model_name"] + "_") :]] = v
                key_to_remove.append(k)

        return self.model_construct_env(**sd_config_attrs), {k: v for k, v in attrs.items() if k not in key_to_remove}

    @classmethod
    @t.overload
    def to_click_options(
        cls, f: t.Callable[..., onediffusion.SDConfig]
    ) -> F[P, ClickFunctionWrapper[..., onediffusion.SDConfig]]:
        ...

    @classmethod
    @t.overload
    def to_click_options(cls, f: t.Callable[P, O_co]) -> F[P, ClickFunctionWrapper[P, O_co]]:
        ...

    @classmethod
    def to_click_options(cls, f: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        """
        Convert current model to click options. This can be used as a decorator for click commands.
        Note that the identifier for all SDConfig will be prefixed with '<model_name>_*', and the generation config
        will be prefixed with '<model_name>_generation_*'.
        """

        for name, field in attr.fields_dict(cls).items():
            ty = cls.__sdserver_hints__.get(name)
            f = dantic.attrs_to_options(name, field, cls.__sdserver_model_name__, typ=ty)(f)

        return cog.optgroup.group(f"{cls.__name__} options")(f)


bentoml_cattr.register_unstructure_hook_factory(
    lambda cls: lenient_issubclass(cls, SDConfig),
    lambda cls: make_dict_unstructure_fn(cls, bentoml_cattr, _cattrs_omit_if_default=False),
)


def structure_sd_config(data: DictStrAny, cls: type[SDConfig]) -> SDConfig:
    """
    Structure a dictionary to a SDConfig object.

    Essentially, if the given dictionary contains a 'generation_config' key, then we will
    use it for SDConfig.generation_config

    Otherwise, we will filter out all keys are first in SDConfig, parse it in, then
    parse the remaining keys into SDConfig.generation_config
    """
    if not LazyType(DictStrAny).isinstance(data):
        raise RuntimeError(f"Expected a dictionary, but got {type(data)}")

    cls_attrs = {
        k: v for k, v in data.items() if k in cls.__sdserver_accepted_keys__
    }
    not_extras = list(cls_attrs)
    # The rest should be passed to extras
    data = {k: v for k, v in data.items() if k not in not_extras}

    return cls(__sdserver_extras__=data, **cls_attrs)


bentoml_cattr.register_structure_hook_func(lambda cls: lenient_issubclass(cls, SDConfig), structure_sd_config)



import os
import typing as t

import torch
from pydantic import BaseModel
import inflection

import bentoml
from bentoml.io import Image, JSON, Multipart
import bentoml.diffusers_simple

import onediffusion

model = os.environ.get("ONEDIFFUSION_MODEL", "{__model_name__}")  # model name
model_name = inflection.underscore(model)
model_id = os.environ.get("ONEDIFFUSION_MODEL_ID", "{__model_id__}")  # model id
pipeline = os.environ.get("ONEDIFFUSION_PIPELINE", "{__pipeline__}")  # pipeline
lora_weights = os.environ.get("ONEDIFFUSION_LORA_WEIGHTS", "{__lora_weights__}")
lora_dir = os.environ.get("ONEDIFFUSION_LORA_DIR", "{__lora_dir__}")

sd_config = onediffusion.AutoConfig.for_model(model)

model_name = inflection.underscore(model)
module = getattr(bentoml.diffusers_simple, model_name)
short_name = module.MODEL_SHORT_NAME

model_runner = module.create_runner(
    model_id=model_id,
    pipeline_class=pipeline,
    lora_weights=lora_weights,
    lora_dir=lora_dir,
)

input_spec_mapping = {}

class Text2ImgArgs(BaseModel):
    prompt: str
    negative_prompt: t.Optional[str] = None
    height: t.Optional[int] = module.DEFAULT_SIZE[0]
    width: t.Optional[int] = module.DEFAULT_SIZE[1]
    num_inference_steps: t.Optional[int] = 50
    guidance_scale: t.Optional[float] = 7.5
    eta: t.Optional[float] = 0.0
    lora_weights: t.Optional[str] = None

    class Config:
        extra = "allow"

input_spec_mapping["text2img"] = JSON.from_sample(
    Text2ImgArgs(prompt="a bento box")
)

class Img2ImgArgs(BaseModel):
    prompt: str
    negative_prompt: t.Optional[str] = None
    num_inference_steps: t.Optional[int] = 50
    guidance_scale: t.Optional[float] = 7.5
    eta: t.Optional[float] = 0.0
    lora_weights: t.Optional[str] = None
    strength: t.Optional[float] = 0.8

input_spec_mapping["img2img"] = Multipart(
    img=Image(),
    data=JSON(pydantic_model=Img2ImgArgs),
)


svc = bentoml.Service(f"onediffusion-{short_name}-service", runners=[model_runner])

endpoint_mapping = {}

def text2img(input_data):
    kwargs = input_data.dict()
    res = model_runner.run(**kwargs)
    images = res[0]
    return images[0]

endpoint_mapping["text2img"] = text2img

def img2img(img, data):
    kwargs = data.dict()
    kwargs["image"] = img
    res = model_runner.run(**kwargs)
    images = res[0]
    return images[0]

endpoint_mapping["img2img"] = img2img

svc.api(input=input_spec_mapping[pipeline], output=Image())(endpoint_mapping[pipeline])



# Copyright 2023 BentoML Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Any build-related utilities. This is used for CI.
"""
from __future__ import annotations

import importlib.metadata
import logging
import os
import typing as t
from pathlib import Path

import fs
import fs.copy
import inflection

import bentoml
import onediffusion
from bentoml._internal.bento.build_config import DockerOptions
from bentoml._internal.bento.build_config import PythonOptions
from bentoml._internal.configuration import get_debug_mode
from bentoml._internal.frameworks.diffusers_runners.utils import get_model_or_download

from .utils import ModelEnv
from .utils import first_not_none
from .utils import is_torch_available
from .utils import pkg


if t.TYPE_CHECKING:
    from fs.base import FS

    from .models.auto.factory import _BaseAutoSDClass

logger = logging.getLogger(__name__)

ONEDIFFUSION_DEV_BUILD = "ONEDIFFUSION_DEV_BUILD"


def build_editable(path: str) -> str | None:
    """Build OneDiffusion if the ONEDIFFUSION_DEV_BUILD environment variable is set."""
    if str(os.environ.get(ONEDIFFUSION_DEV_BUILD, False)).lower() != "true":
        return

    # We need to build the package in editable mode, so that we can import it
    from build import ProjectBuilder
    from build.env import IsolatedEnvBuilder

    module_location = pkg.source_locations("onediffusion")
    if not module_location:
        raise RuntimeError(
            "Could not find the source location of OneDiffusion. Make sure to unset"
            " ONEDIFFUSION_DEV_BUILD if you are developing OneDiffusion."
        )
    pyproject_path = Path(module_location).parent.parent / "pyproject.toml"
    if os.path.isfile(pyproject_path.__fspath__()):
        logger.info("OneDiffusion is installed in editable mode. Generating built wheels...")
        with IsolatedEnvBuilder() as env:
            builder = ProjectBuilder(pyproject_path.parent)
            builder.python_executable = env.executable
            builder.scripts_dir = env.scripts_dir
            env.install(builder.build_system_requires)
            return builder.build("wheel", path, config_settings={"--global-option": "--quiet"})
    raise RuntimeError(
        "Custom OneDiffusion build is currently not supported. Please install OneDiffusion from PyPI or built it from Git source."
    )


def construct_python_options(
    sd: onediffusion.SD[t.Any, t.Any],
    sd_fs: FS,
    extra_dependencies: tuple[str, ...] | None = None,
) -> PythonOptions:
    packages = ["onediffusion"]
    # NOTE: add onediffusion to the default dependencies
    # if users has onediffusion custom built wheels, it will still respect
    # that since bentoml will always install dependencies from requirements.txt
    # first, then proceed to install everything inside the wheels/ folder.
    if extra_dependencies is not None:
        packages += [f"onediffusion[{k}]" for k in extra_dependencies]

    if sd.config["requirements"] is not None:
        packages.extend(sd.config["requirements"])

    if not (str(os.environ.get("BENTOML_BUNDLE_LOCAL_BUILD", False)).lower() == "false"):
        packages.append(f"bentoml>={'.'.join([str(i) for i in pkg.pkg_version_info('bentoml')])}")

    env: ModelEnv = sd.config["env"]
    framework_envvar = env["framework_value"]
    assert is_torch_available(), "PyTorch is not available. Make sure to have it locally installed."
    packages.extend([f"torch>={importlib.metadata.version('torch')}"])

    wheels: list[str] = []
    built_wheels = build_editable(sd_fs.getsyspath("/"))
    if built_wheels is not None:
        wheels.append(sd_fs.getsyspath(f"/{built_wheels.split('/')[-1]}"))

    return PythonOptions(packages=packages, wheels=wheels, lock_packages=True)


def construct_docker_options(
    sd: onediffusion.SD[t.Any, t.Any],
    _: FS,
    workers_per_resource: int | float,
) -> DockerOptions:
    _bentoml_config_options = os.environ.pop("BENTOML_CONFIG_OPTIONS", "")
    _bentoml_config_options_opts = [
        "api_server.traffic.timeout=36000",  # NOTE: Currently we hardcode this value
        f'runners."sd-{sd.config["start_name"]}-runner".traffic.timeout={sd.config["timeout"]}',
        f'runners."sd-{sd.config["start_name"]}-runner".workers_per_resource={workers_per_resource}',
    ]
    _bentoml_config_options += " " if _bentoml_config_options else "" + " ".join(_bentoml_config_options_opts)
    env: ModelEnv = sd.config["env"]

    env_dict = {
        env.framework: env.framework_value,
        env.config: f"'{sd.config.model_dump_json().decode()}'",
        "ONEDIFFUSION_MODEL": sd.config["model_name"],
        "ONEDIFFUSION_MODEL_ID": sd.model_id,
        "ONEDIFFUSION_PIPELINE": sd.pipeline,
        "BENTOML_DEBUG": str(get_debug_mode()),
        "BENTOML_CONFIG_OPTIONS": _bentoml_config_options,
    }

    # NOTE: Torch 2.0 currently only support 11.6 as the latest CUDA version
    return DockerOptions(cuda_version="11.6", env=env_dict, system_packages=["git"])


@t.overload
def build(
    model_name: str,
    *,
    model_id: str | None = ...,
    pipeline: str | None = ...,
    _extra_dependencies: tuple[str, ...] | None = ...,
    _workers_per_resource: int | float | None = ...,
    _overwrite_existing_bento: bool = ...,
    __cli__: t.Literal[False] = ...,
    **attrs: t.Any,
) -> bentoml.Bento:
    ...


@t.overload
def build(
    model_name: str,
    *,
    model_id: str | None = ...,
    pipeline: str | None = ...,
    _extra_dependencies: tuple[str, ...] | None = ...,
    _workers_per_resource: int | float | None = ...,
    _overwrite_existing_bento: bool = ...,
    __cli__: t.Literal[True] = ...,
    **attrs: t.Any,
) -> tuple[bentoml.Bento, bool]:
    ...


def _build_bento(
    bento_tag: bentoml.Tag,
    service_name: str,
    sd_fs: FS,
    sd: onediffusion.SD[t.Any, t.Any],
    workers_per_resource: int | float,
    extra_dependencies: tuple[str, ...] | None = None,
) -> bentoml.Bento:
    framework_envvar = sd.config["env"]["framework_value"]
    labels = {}
    labels.update({"_type": sd.sd_type, "_framework": framework_envvar})
    logger.info("Building Bento for diffusion model '%s'", sd.config["start_name"])

    return bentoml.bentos.build(
        f"{service_name}:svc",
        name=bento_tag.name,
        labels=labels,
        description=f"OneDiffusion service for {sd.config['start_name']}",
        exclude=["/venv", "__pycache__/", "*.py[cod]", "*$py.class"],
        python=construct_python_options(sd, sd_fs, extra_dependencies),
        docker=construct_docker_options(sd, sd_fs, workers_per_resource),
        version=bento_tag.version,
        build_ctx=sd_fs.getsyspath("/"),
    )


def build(
    model_name: str,
    *,
    model_id: str | None = None,
    pipeline: str | None = None,
    lora_weights: str | None = None,
    lora_dir: str | None = None,
    bento_name: str | None = None,
    bento_version: str | None = None,
    _extra_dependencies: tuple[str, ...] | None = None,
    _workers_per_resource: int | float | None = None,
    _overwrite_existing_bento: bool = False,
    __cli__: bool = False,
    **attrs: t.Any,
) -> tuple[bentoml.Bento, bool] | bentoml.Bento:
    """Package a diffusion model into a Bento.

    The diffusion model will be built into a BentoService with the following structure:

    Other parameters including model_name, model_id, pipeline and attrs will be passed to the OneDiffusion class itself.
    """

    _previously_built = False
    current_model_envvar = os.environ.pop("ONEDIFFUSION_MODEL", None)
    current_model_id_envvar = os.environ.pop("ONEDIFFUSION_MODEL_ID", None)

    sd_config = onediffusion.AutoConfig.for_model(model_name)

    logger.info("Packing '%s' into a Bento with kwargs=%s...", model_name, attrs)

    # NOTE: We set this environment variable so that our service.py logic won't raise RuntimeError
    # during build. This is a current limitation of bentoml build where we actually import the service.py into sys.path
    try:
        os.environ["ONEDIFFUSION_MODEL"] = inflection.underscore(model_name)

        framework_envvar = sd_config["env"]["framework_value"]
        sd = t.cast(
            "_BaseAutoSDClass",
            onediffusion[framework_envvar],  # type: ignore (internal API)
        ).for_model(
            model_name,
            model_id=model_id,
            pipeline=pipeline,
            sd_config=sd_config,
            **attrs,
        )

        os.environ["ONEDIFFUSION_MODEL_ID"] = sd.model_id

        labels = {}
        labels.update({"_type": sd.sd_type, "_framework": framework_envvar})
        target_service_name = f"generated_{sd_config['model_name']}_service.py"
        workers_per_resource = first_not_none(_workers_per_resource, default=sd_config["workers_per_resource"])

        with fs.open_fs(f"temp://od_{sd_config['model_name']}") as sd_fs:

            _model_name = inflection.underscore(model_name)

            # add service.py definition to this temporary folder
            base_dir = os.path.dirname(__file__)
            src_service_path = os.path.join(base_dir, "models", _model_name, "service.py")
            target_service_name = f"generated_{sd_config['model_name']}_service.py"

            with open(src_service_path) as f:
                import bentoml.diffusers_simple
                module = getattr(bentoml.diffusers_simple, _model_name)
                short_name = module.MODEL_SHORT_NAME
                _model_id = model_id or module.DEFAULT_MODEL_ID
                bento_model = get_model_or_download(
                    model_name=short_name,
                    model_id=_model_id,
                )
                model_version = bento_model.tag.version

                lora_files_to_copy: list[str] = []

                content = f.read()
                content = content.replace(r"{__model_name__}", model_name)
                content = content.replace(r"{__model_id__}", sd.model_id)
                content = content.replace(r"{__pipeline__}", sd.pipeline)
                if lora_weights:
                    lora_path = os.path.realpath(os.path.expanduser(lora_weights))
                    assert os.path.isfile(lora_path), "Currently OneDiffusion only support package local lora weights into bento"
                    lora_filename = os.path.basename(lora_path)
                    content = content.replace(r"{__lora_weights__}", lora_filename)
                    lora_files_to_copy.append(lora_path)
                else:
                    content = content.replace(r"{__lora_weights__}", "")
                if lora_weights or lora_dir:
                    content = content.replace(r"{__lora_dir__}", "/home/bentoml/bento/src/lora")
                else:
                    content = content.replace(r"{__lora_dir__}", "")
                sd_fs.writetext(target_service_name, content)

                # package lora files
                if lora_dir:
                    lora_dir_path = os.path.realpath(os.path.expanduser(lora_dir))
                    assert os.path.isdir(lora_dir_path), "Currently OneDiffusion only support package local lora dir into bent"

                    with fs.open_fs(lora_dir_path) as lora_fs:
                        fs.copy.copy_dir(lora_fs, "", sd_fs, "lora")

                if lora_files_to_copy:
                    sd_fs.makedir("lora", recreate=True)

                    for path in lora_files_to_copy:
                        dir_path = os.path.dirname(path)
                        file_name = os.path.basename(path)
                        target_path = os.path.join("lora", file_name)
                        with fs.open_fs(dir_path) as src_fs:
                            fs.copy.copy_file(src_fs, file_name, sd_fs, target_path)

            name = bento_name or f"{framework_envvar}-{sd.sd_type}-{sd.pipeline}"
            version = bento_version or model_version
            bento_tag = bentoml.Tag(name=name, version=version)
            try:
                bento = bentoml.get(bento_tag)
                if _overwrite_existing_bento:
                    logger.info("Overwriting previously saved Bento.")
                    bentoml.delete(bento_tag)
                    bento = _build_bento(
                        bento_tag,
                        target_service_name,
                        sd_fs,
                        sd,
                        workers_per_resource=workers_per_resource,
                        extra_dependencies=_extra_dependencies,
                    )
                _previously_built = True
            except bentoml.exceptions.NotFound:
                logger.info("Building Bento for diffusion model '%s'", sd_config["start_name"])
                bento = _build_bento(
                    bento_tag,
                    target_service_name,
                    sd_fs,
                    sd,
                    workers_per_resource=workers_per_resource,
                    extra_dependencies=_extra_dependencies,
                )
            return (bento, _previously_built) if __cli__ else bento
    except Exception as e:
        logger.error("\nException caught during building diffusion model %s: \n", model_name, exc_info=e)
        raise
    finally:
        del os.environ["ONEDIFFUSION_MODEL"]
        del os.environ["ONEDIFFUSION_MODEL_ID"]
        # restore original ONEDIFFUSION_MODEL envvar if set.
        if current_model_envvar is not None:
            os.environ["ONEDIFFUSION_MODEL"] = current_model_envvar
        if current_model_id_envvar is not None:
            os.environ["ONEDIFFUSION_MODEL_ID"] = current_model_id_envvar
